#!/usr/bin/python
# dht_service.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (dht_service.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
"""
..

module:: dht_service
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
from __future__ import print_function

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

import os
import sys
import hashlib
import random
import optparse
import pprint
import json

#------------------------------------------------------------------------------

from twisted.internet import reactor  # @UnresolvedImport
from twisted.internet.task import LoopingCall  #@UnresolvedImport
from twisted.internet.defer import Deferred, DeferredList, fail  # @UnresolvedImport

#------------------------------------------------------------------------------

if __name__ == '__main__':
    import os.path as _p
    sys.path.insert(0, _p.abspath(_p.join(_p.dirname(_p.abspath(sys.argv[0])), '..')))

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.system import bpio
from bitdust.system import local_fs

from bitdust.main import settings
from bitdust.main import events

from bitdust.lib import strng
from bitdust.lib import utime
from bitdust.lib import jsn

from bitdust.userid import id_url

from bitdust.dht import known_nodes

#------------------------------------------------------------------------------

from bitdust_forks.entangled.kademlia import constants  # @UnresolvedImport
from bitdust_forks.entangled.kademlia.datastore import SQLiteVersionedJsonDataStore  # @UnresolvedImport
from bitdust_forks.entangled.kademlia.node import rpcmethod, MultiLayerNode  # @UnresolvedImport
from bitdust_forks.entangled.kademlia.protocol import KademliaMultiLayerProtocol, encoding, msgformat  # @UnresolvedImport
from bitdust_forks.entangled.kademlia.routingtable import TreeRoutingTable  # @UnresolvedImport
from bitdust_forks.entangled.kademlia.contact import Contact  # @UnresolvedImport

#------------------------------------------------------------------------------

KEY_EXPIRE_MIN_SECONDS = 60*2
KEY_EXPIRE_MAX_SECONDS = constants.dataExpireSecondsDefaut
RECEIVING_FREQUENCY_SEC = 0.01
SENDING_FREQUENCY_SEC = 0.02  # must be always slower than receiving frequency!
RECEIVING_QUEUE_LENGTH_CRITICAL = 100
SENDING_QUEUE_LENGTH_CRITICAL = 50
DEFAULT_CACHE_TTL = 60*60*3

#------------------------------------------------------------------------------

_MyNode = None
_ActiveLookup = None
_ActiveLookupLayerID = None
_Counters = {}
_ProtocolVersion = 7
_Cache = {}

#------------------------------------------------------------------------------


def init(udp_port, dht_dir_path=None, open_layers=[]):
    global _MyNode
    if _MyNode is not None:
        if _Debug:
            lg.out(_DebugLevel, 'dht_service.init SKIP, DHTNode already exist')
        return
    nw_info = known_nodes.network_info()
    if 'bucket_size' in nw_info:
        constants.k = nw_info['bucket_size']
    if 'parallel_calls' in nw_info:
        constants.alpha = nw_info['parallel_calls']
    if 'rpc_timeout' in nw_info:
        constants.rpcTimeout = nw_info['rpc_timeout']
    if 'refresh_timeout' in nw_info:
        constants.refreshTimeout = nw_info['refresh_timeout']
    if 'max_age' in nw_info:
        constants.dataExpireTimeout = nw_info['max_age']
    if 'default_age' in nw_info:
        constants.dataExpireSecondsDefaut = nw_info['default_age']
    if dht_dir_path is None:
        dht_dir_path = settings.ServiceDir('service_entangled_dht')
    if not os.path.isdir(dht_dir_path):
        os.makedirs(dht_dir_path)
    list_layers = []
    if os.path.isdir(dht_dir_path):
        list_layers = os.listdir(dht_dir_path)
    cache_dir_path = os.path.join(dht_dir_path, 'cache')
    if not os.path.isdir(cache_dir_path):
        os.makedirs(cache_dir_path)
    load_cache(cache_dir_path)
    if _Debug:
        lg.dbg(_DebugLevel, 'dht_dir_path=%r list_layers=%r network_info=%r' % (dht_dir_path, list_layers, nw_info))
    layerStores = {}
    for layer_filename in list_layers:
        if not layer_filename.startswith('db_'):
            continue
        try:
            layer_name = layer_filename.replace('db_', '')
            layer_id = int(layer_name)
        except:
            continue
        db_file_path = os.path.join(dht_dir_path, 'db_%d' % layer_id)
        dbPath = bpio.portablePath(db_file_path)
        if _Debug:
            lg.dbg(_DebugLevel, 'found existing db layer: %r at %r' % (layer_id, dbPath))
        try:
            dataStore = SQLiteVersionedJsonDataStore(dbFile=dbPath)
        except:
            lg.warn('failed reading DHT records, removing %s and starting clean DB' % dbPath)
            lg.exc()
            try:
                os.remove(dbPath)
            except:
                pass
            dataStore = SQLiteVersionedJsonDataStore(dbFile=dbPath)
        layerStores[layer_id] = dataStore
    if not layerStores:
        db_file_path = os.path.join(dht_dir_path, 'db_0')
        dbPath = bpio.portablePath(db_file_path)
        layerStores[0] = SQLiteVersionedJsonDataStore(dbFile=dbPath)
    _MyNode = DHTNode(
        udpPort=udp_port,
        dataStores=layerStores,
        networkProtocol=DHTProtocol,
    )
    for layer_id in open_layers:
        open_layer(layer_id=layer_id, dht_dir_path=dht_dir_path, connect_now=False)
    if _Debug:
        lg.out(_DebugLevel, 'dht_service.init UDP port is %d, DB file path is %s, my DHT ID is %s' % (udp_port, dht_dir_path, _MyNode.layers[0]))


def shutdown():
    global _MyNode
    if _MyNode is not None:
        for ds in _MyNode._dataStores.values():
            ds._db.close()
        _MyNode._protocol.node = None
        del _MyNode
        _MyNode = None
        if _Debug:
            lg.out(_DebugLevel, 'dht_service.shutdown')
    else:
        lg.warn('DHTNode not exist')


#------------------------------------------------------------------------------


def node():
    global _MyNode
    return _MyNode


def connect(seed_nodes=[], layer_id=0, attach=False):
    global _MyNode
    result = Deferred()

    if _Debug:
        lg.args(_DebugLevel, layer_id=layer_id, attach=attach, seed_nodes=seed_nodes)

    if not node():
        lg.err('node is not initialized')
        result.errback(Exception('node is not initialized'))
        return result

    joinDeferred = node().connectingTask(layer_id)
    if joinDeferred:
        if _Debug:
            lg.out(_DebugLevel, 'dht_service.connect SKIP, already joining layer %d' % layer_id)
        if joinDeferred.called:
            lg.warn('joinDeferred already called')
            reactor.callLater(0, result.callback, True)  # @UndefinedVariable
        else:
            joinDeferred.addBoth(lambda x: result.callback(True))
        return result

    if node().refreshers.get(layer_id, None) and node().refreshers[layer_id].active():
        node().refreshers[layer_id].reset(0)
        if _Debug:
            lg.out(_DebugLevel, 'dht_service.connect SKIP seems like DHT already active, RESET current refresher task')
        reactor.callLater(0, result.callback, True)  # @UndefinedVariable
        return result

    if not node().listener:
        node().listenUDP()
        if _Debug:
            lg.out(_DebugLevel, 'dht_service.connect opened a new listener : %r' % node().listener)

    if _Debug:
        lg.out(_DebugLevel, 'dht_service.connect STARTING with %d known nodes in layer %d:' % (len(seed_nodes), layer_id))
        for onenode in seed_nodes:
            lg.out(_DebugLevel, '    %s:%s' % onenode)

    if not seed_nodes:
        if _Debug:
            lg.out(_DebugLevel, 'dht_service.connect  SKIP : no seed nodes provided')
        reactor.callLater(0, result.callback, True)  # @UndefinedVariable
        events.send('dht-layer-connected', data=dict(layer_id=layer_id))
        return result

    def _on_connected(ok):
        if _Debug:
            lg.args(_DebugLevel, ok=ok, layer_id=layer_id)
        events.send('dht-layer-connected', data=dict(layer_id=layer_id))
        return ok

    result.addCallback(_on_connected)

    def _on_join_success(live_contacts, resolved_seed_nodes, _layer_id):
        ok = live_contacts and len(live_contacts) > 0 and live_contacts[0]
        if isinstance(live_contacts, dict):
            lg.err('Unexpected result from joinNetwork: %s' % pprint.pformat(live_contacts))
        else:
            if ok:
                if _Debug:
                    lg.out(_DebugLevel, 'dht_service.connect DHT JOIN SUCCESS   layer_id=%d attach=%r' % (_layer_id, attach))
            else:
                lg.warn('No live DHT contacts found...  your node is NOT CONNECTED TO DHT NETWORK at layer %d' % _layer_id)
        if _Debug:
            lg.out(_DebugLevel, 'for layer %d found DHT nodes: %s' % (_layer_id, live_contacts))
            lg.out(_DebugLevel, 'resolved SEED nodes: %r' % resolved_seed_nodes)
            lg.out(_DebugLevel, 'DHT node is active, ID%d=[%s]' % (_layer_id, node().layers[_layer_id]))
        reactor.callLater(0, result.callback, ok)  # @UndefinedVariable
        return live_contacts

    def _on_join_failed(x):
        if _Debug:
            lg.out(_DebugLevel, 'dht_service.connect DHT JOIN FAILED : %s' % x)
        result.errback(x)
        return None

    def _on_hosts_resolved(resolved_seed_nodes, _layer_id):
        if _Debug:
            lg.out(_DebugLevel, 'dht_service.connect RESOLVED %d live nodes' % (len(resolved_seed_nodes)))
            for onenode in resolved_seed_nodes:
                lg.out(_DebugLevel, '    %s:%s' % onenode)
        if not resolved_seed_nodes:
            resolved_seed_nodes = []
        d = node().joinNetwork(resolved_seed_nodes, layerID=_layer_id, attach=attach)
        d.addCallback(_on_join_success, resolved_seed_nodes, _layer_id=_layer_id)
        d.addErrback(_on_join_failed)
        if not node().expire_task.running:
            # reactor.callLater(random.randint(0, 60), node().expire_task.start, int(KEY_EXPIRE_MIN_SECONDS / 2), now=True)  # @UndefinedVariable
            node().expire_task.start(int(KEY_EXPIRE_MIN_SECONDS/2), now=True)  # @UndefinedVariable
        return resolved_seed_nodes

    def _on_hosts_resolve_failed(x):
        if _Debug:
            lg.out(_DebugLevel, 'dht_service.connect ERROR : hosts not resolved: %s' % x)
        result.errback(x)
        return None

    d = resolve_hosts(seed_nodes)
    d.addCallback(_on_hosts_resolved, _layer_id=layer_id)
    d.addErrback(_on_hosts_resolve_failed)
    return result


def suspend(layer_id):
    if not node():
        return False
    if _Debug:
        lg.out(_DebugLevel, 'dht_service.suspend    layer %d' % layer_id)
    node().leaveNetwork(layer_id)
    return True


def disconnect():
    global _MyNode
    if not node():
        return False
    if node().expire_task and node().expire_task.running:
        node().expire_task.stop()
    for refresher in node().refreshers.values():
        if refresher and refresher.active():
            refresher.cancel()
    node().refreshers.clear()
    if node().listener:
        node().listener.stopListening()
    return True


def reconnect():
    global _MyNode
    if not node():
        return None
    if _Debug:
        lg.out(_DebugLevel, 'dht_service.reconnect')
    return node().reconnect()


#------------------------------------------------------------------------------


def is_layer_active(layer_id):
    if not node():
        return False
    return layer_id in node().active_layers


def open_layer(layer_id, seed_nodes=[], dht_dir_path=None, connect_now=False, attach=False):
    global _MyNode
    if not node():
        result = Deferred()
        result.callback(False)
        return result
    if not layer_id in node().layers:
        if dht_dir_path is None:
            dht_dir_path = settings.ServiceDir('service_entangled_dht')
        if not os.path.isdir(dht_dir_path):
            os.makedirs(dht_dir_path)
        layer_file_path = os.path.join(dht_dir_path, 'db_%d' % layer_id)
        dbPath = bpio.portablePath(layer_file_path)
        if not node().createLayer(layer_id, dataStore=SQLiteVersionedJsonDataStore(dbFile=dbPath)):
            lg.warn('failed to create DHT layer %d' % layer_id)
            result = Deferred()
            result.callback(False)
            return result
    if not connect_now:
        result = Deferred()
        result.callback(True)
        return result
    result = connect(seed_nodes=seed_nodes, layer_id=layer_id, attach=attach)
    lg.info('DHT layer %d opened and connecting to the seed nodes: %r' % (layer_id, seed_nodes))
    return result


def close_layer(layer_id):
    global _MyNode
    if not node():
        return False
    lg.info('destroying DHT layer %d' % layer_id)
    return node().destroyLayer(layer_id)


#------------------------------------------------------------------------------


def count(name):
    global _Counters
    if name not in _Counters:
        _Counters[name] = 0
    _Counters[name] += 1
    return True


def counter(name):
    global _Counters
    return _Counters.get(name, 0)


def drop_counters():
    global _Counters
    ret = _Counters.copy()
    _Counters.clear()
    return ret


#------------------------------------------------------------------------------


def on_host_resolved(ip, port, host, result_list, total_hosts, result_defer):
    if not strng.is_string(ip) or port is None:
        result_list.append(None)
        lg.warn('%r failed to resolve' % host)
    else:
        result_list.append((
            ip,
            port,
        ))
    if len(result_list) != total_hosts:
        return None
    return result_defer.callback([_f for _f in result_list if _f])


def on_host_failed(err, host, result_list, total_hosts, result_defer):
    lg.warn('%r failed to resolve: %r' % (host, err))
    result_list.append(None)
    if len(result_list) != total_hosts:
        return None
    return result_defer.callback([_f for _f in result_list if _f])


def resolve_hosts(nodes_list):
    result_defer = Deferred()
    if not nodes_list:
        result_defer.callback([])
        return result_defer
    result_list = []
    for node_tuple in nodes_list:
        d = reactor.resolve(strng.to_text(node_tuple[0]))  #@UndefinedVariable
        d.addCallback(on_host_resolved, node_tuple[1], node_tuple[0], result_list, len(nodes_list), result_defer)
        d.addErrback(on_host_failed, node_tuple[0], result_list, len(nodes_list), result_defer)
    return result_defer


#------------------------------------------------------------------------------


def random_key():
    return key_to_hash(strng.to_text(random.getrandbits(255)).encode())


def key_to_hash(key):
    key = strng.to_bin(key)
    h = hashlib.sha1()
    h.update(key)
    return h.hexdigest()


#------------------------------------------------------------------------------


def make_key(key, prefix, index=0, version=None):
    global _ProtocolVersion
    if not version:
        version = _ProtocolVersion
    return '{}:{}:{}:{}'.format(prefix, key, index, version)


def split_key(key_str):
    key_str = strng.to_text(key_str)
    prefix, key, index, version = key_str.split(':')
    return dict(
        key=key,
        prefix=prefix,
        index=index,
        version=version,
    )


#------------------------------------------------------------------------------


def on_success(result, method, key, *args, **kwargs):
    if _Debug:
        lg.out(_DebugLevel, 'dht_service.on_success   %s(%r)    result is %s' % (method, key, type(result)))
    return result


def on_error(x, method, key):
    try:
        errmsg = x.value.subFailure.getErrorMessage()
    except:
        try:
            errmsg = x.getErrorMessage()
        except:
            try:
                errmsg = x.value
            except:
                try:
                    errmsg = str(x)
                except:
                    errmsg = 'Unknown Error'
    if _Debug:
        lg.out(_DebugLevel, 'dht_service.on_error   %s(%s)   returned an ERROR:\n%r' % (method, key, errmsg))
    return x


#------------------------------------------------------------------------------


def get_value(key, layer_id=0, parallel_calls=None):
    if not node():
        return fail(Exception('DHT service is off'))
    count('get_value_%s' % key)
    key_hash = key_to_hash(key)
    if _Debug:
        lg.out(_DebugLevel, 'dht_service.get_value key=[%r] key_hash=%s' % (key, key_hash))
    d = node().iterativeFindValue(
        key=key_hash,
        rpc='findValue',
        layerID=layer_id,
        parallel_calls=parallel_calls,
    )
    d.addCallback(on_success, 'get_value', key)
    d.addErrback(on_error, 'get_value', key)
    return d


def set_value(key, value, age=0, expire=KEY_EXPIRE_MAX_SECONDS, collect_results=True, layer_id=0, parallel_calls=None):
    if not node():
        return fail(Exception('DHT service is off'))
    count('set_value_%s' % key)
    sz_bytes = len(value)
    key_hash = key_to_hash(key)
    if _Debug:
        lg.out(_DebugLevel, 'dht_service.set_value key=[%s] key_hash=%s with %d bytes for %d seconds' % (key, key_hash, sz_bytes, expire))
    if expire < KEY_EXPIRE_MIN_SECONDS:
        expire = KEY_EXPIRE_MIN_SECONDS
    if expire > KEY_EXPIRE_MAX_SECONDS:
        expire = KEY_EXPIRE_MAX_SECONDS
    d = node().iterativeStore(
        key=key_hash,
        value=value,
        age=age,
        expireSeconds=expire,
        collect_results=collect_results,
        layerID=layer_id,
        parallel_calls=parallel_calls,
    )
    d.addCallback(on_success, 'set_value', key, value)
    d.addErrback(on_error, 'set_value', key)
    return d


def delete_key(key, layer_id=0, parallel_calls=None):
    if not node():
        return fail(Exception('DHT service is off'))
    count('delete_key_%s' % key)
    key_hash = key_to_hash(key)
    if _Debug:
        lg.out(_DebugLevel, 'dht_service.delete_key [%s] key_hash=%s' % (key, key_hash))
    d = node().iterativeDelete(key_hash, layerID=layer_id, parallel_calls=parallel_calls)
    d.addCallback(on_success, 'delete_value', key)
    d.addErrback(on_error, 'delete_key', key)
    return d


#------------------------------------------------------------------------------


def on_read_json_response(response, key, result_defer=None):
    if _Debug:
        lg.out(_DebugLevel + 6, 'dht_service.on_read_json_response [%r] : %s' % (key, type(response)))
    json_value = None
    if isinstance(response, list):
        if _Debug:
            lg.out(_DebugLevel, '        response is a list, value not found')
        if result_defer:
            result_defer.callback(response)
        return None
    if isinstance(response, dict):
        if response.get('values'):
            try:
                latest_revision = 0
                latest = 0
                json_value = jsn.loads_text(response['values'][0][0])
                for v in response['values']:
                    j = jsn.loads_text(v[0])
                    rev = j.get('revision', -1)
                    if rev >= 0:
                        if rev > latest_revision:
                            latest = v[1]
                            latest_revision = rev
                            json_value = j
                    else:
                        if v[1] > latest:
                            latest = v[1]
                            latest_revision = rev
                            json_value = j
            except:
                lg.exc()
                if _Debug:
                    lg.out(_DebugLevel, '        invalid json value found in DHT, return None')
                if result_defer:
                    result_defer.errback(Exception('invalid json value found in DHT'))
                return None
        else:
            if _Debug:
                lg.out(_DebugLevel, '        response is a dict, "values" field is empty, value not found')
            if result_defer:
                result_defer.callback(response.get('activeContacts', []))
            return None
    if _Debug:
        lg.out(_DebugLevel, '        response is a dict, value is OK')
    if result_defer:
        result_defer.callback(json_value)
    return json_value


def get_json_value(key, layer_id=0, update_cache=True):
    if _Debug:
        lg.out(_DebugLevel, 'dht_service.get_json_value key=[%r] layer_id=%d update_cache=%s' % (key, layer_id, update_cache))
    ret = Deferred()
    d = get_value(key, layer_id=layer_id)
    d.addCallback(on_read_json_response, key, ret)
    d.addErrback(ret.errback)
    if update_cache:
        d.addCallback(on_json_response_to_be_cached, key=key, layer_id=layer_id)
    return ret


def set_json_value(key, json_data, age=0, expire=KEY_EXPIRE_MAX_SECONDS, collect_results=True, layer_id=0):
    if not node():
        return fail(Exception('DHT service is off'))
    try:
        value = jsn.dumps(json_data, indent=0, sort_keys=True, separators=(',', ':'))
    except:
        return fail(Exception('bad input json data'))
    if _Debug:
        lg.out(_DebugLevel, 'dht_service.set_json_value key=[%r] layer_id=%d with %d bytes' % (key, layer_id, len(repr(value))))
    return set_value(key=key, value=value, age=age, expire=expire, collect_results=collect_results, layer_id=layer_id)


#------------------------------------------------------------------------------


def validate_rules(value, key, rules, result_defer=None, raise_for_result=False, populate_meta_fields=False):
    """
    Will be executed on both sides: sender and receiver for each (key,value) get / set request on DHT.
    Can return in `result_defer` : errback, callback with list (closest nodes), callback with dict
    """
    if not rules:
        lg.err('DHT record must have validation rules applied')
        if result_defer:
            result_defer.errback(Exception('data must have validation rules applied'))
        return None

    if not isinstance(value, dict):
        if _Debug:
            lg.out(_DebugLevel, 'dht_service.validate_rules    key=[%s] not found : %s' % (key, type(value)))
        if not result_defer:
            return value if populate_meta_fields else None
        if raise_for_result:
            result_defer.errback(Exception('value not found'))
            return None
        if not populate_meta_fields:
            result_defer.callback(None)
            return None
        result_defer.callback(value)
        return None

    try:
        expected_record_type = rules['type'][0]['arg']
    except:
        lg.exc()
        if result_defer:
            result_defer.errback(Exception('invalid validation rules can not be applied'))
        return None

    passed = True
    errors = []
    try:
        if populate_meta_fields:
            # Otherwise those records will not pass validation
            if value.get('type') != expected_record_type:
                value['type'] = expected_record_type
            if value.get('key') != key:
                value['key'] = key

        for field, field_rules in rules.items():
            if _Debug:
                lg.out(_DebugLevel, '    %r : %r' % (field, field_rules))
            for rule in field_rules:
                if 'op' not in rule:
                    lg.warn('incorrect validation rule found: %r' % rule)
                    continue
                if rule['op'] == 'equal' and rule.get('arg') != strng.to_text(value.get(field)):
                    passed = False
                    errors.append((
                        field,
                        rule,
                    ))
                    break
                if rule['op'] == 'exist' and field not in value:
                    passed = False
                    errors.append((
                        field,
                        rule,
                    ))
                    break
            if not passed:
                break
    except Exception as exc:
        lg.exc()
        passed = False
        errors = [(
            'unknown',
            exc,
        )]
    if not passed:
        lg.exc(exc_value=Exception('DHT record validation failed, errors: %s' % errors))
        if result_defer:
            result_defer.errback(Exception('DHT record validation failed: %r' % errors))
        return None
    if _Debug:
        lg.out(_DebugLevel, 'dht_service.validate_rules   key=[%s] : value is OK' % key)
    if result_defer:
        result_defer.callback(value)
    return value


def validate_before_store(key, value, originalPublisherID, age, expireSeconds, **kwargs):
    """
    Will be executed on receiver side for each (key,value) set request on DHT
    """
    try:
        json_new_value = jsn.loads(value)
    except:
        # not a json data to be written - this is not valid
        lg.exc()
        raise ValueError('input data is not a json value: %r' % value)
    if _Debug:
        lg.out(_DebugLevel + 8, 'dht_service.validate_before_store key=[%s] with %d bytes value' % (key, len(value)))
    new_record_type = json_new_value.get('type')
    if not new_record_type:
        if _Debug:
            lg.out(_DebugLevel, '        new json data do not have "type" field present, store operation FAILED')
        raise ValueError('input data do not have "type" field present: %r' % json_new_value)
    layer_id = kwargs.get('layerID', 0)
    if key not in node()._dataStores[layer_id]:
        if _Debug:
            lg.out(_DebugLevel, '        previous value not exists yet, store OK')
        return True
    prev_value = node()._dataStores[layer_id][key]
    try:
        json_prev_value = jsn.loads(prev_value)
    except:
        if _Debug:
            lg.out(_DebugLevel, '        current value in DHT is not a json data, will be overwritten, store OK')
        return True
    prev_record_type = json_prev_value.get('type')
    if prev_record_type and prev_record_type != new_record_type:
        if _Debug:
            lg.out(_DebugLevel, '        new json data type did not match to existing record type, store operation FAILED')
        raise ValueError('new json data type do not match to existing record type: %r' % json_prev_value)
    # TODO: need to include "key" field into DHT record and validate it as well
    # new_record_key = json_new_value.get('key')
    # if not new_record_key:
    #     if _Debug:
    #         lg.out(_DebugLevel, '        new json data do not have "key" field present, store operation FAILED')
    #     return False
    # if new_record_key != key:
    #     if _Debug:
    #         lg.out(_DebugLevel, '        new json data do not have "key" field set properly, store operation FAILED')
    #     return False
    # prev_record_key = json_prev_value.get('key')
    # if prev_record_key and prev_record_key != new_record_key:
    #     if _Debug:
    #         lg.out(_DebugLevel, '        new json data "key" field do not match to existing record "key", store operation FAILED')
    #     return False
    try:
        prev_revision = int(json_prev_value['revision'])
    except:
        prev_revision = -1
    try:
        new_revision = int(json_new_value['revision'])
    except:
        new_revision = -1
    if prev_revision >= 0:
        if new_revision < 0:
            if _Debug:
                lg.out(_DebugLevel, '        new json data must have a revision, store operation FAILED')
            raise ValueError('new json data must have a revision')
        if new_revision < prev_revision:
            if _Debug:
                lg.out(_DebugLevel, '        new json data must increment revision number, store operation FAILED')
            raise ValueError('new json data must increment revision number, current revision is %d ' % prev_revision)
        if new_revision == prev_revision:
            if prev_record_type == 'suppliers':
                prev_ecc_map = json_prev_value.get('ecc_map')
                new_ecc_map = json_new_value.get('ecc_map')
                if prev_ecc_map and new_ecc_map != prev_ecc_map:
                    if _Debug:
                        lg.out(_DebugLevel, '        new json data have same revision but different ecc_map, store operation FAILED')
                    raise ValueError('new json data have same revision but different ecc_map, current revision is %d ' % prev_revision)
                prev_suppliers = id_url.to_bin_list(json_prev_value.get('suppliers', []))
                new_suppliers = id_url.to_bin_list(json_new_value.get('suppliers', []))
                if prev_suppliers != new_suppliers:
                    if _Debug:
                        lg.out(_DebugLevel, '        new json data have same revision but different suppliers list, store operation FAILED')
                    raise ValueError('new json data have same revision but different suppliers list, current revision is %d ' % prev_revision)
            if prev_record_type == 'message_broker':
                prev_broker_idurl = json_prev_value.get('broker_idurl')
                new_broker_idurl = json_new_value.get('broker_idurl')
                prev_position = json_prev_value.get('position')
                new_position = json_new_value.get('position')
                if prev_broker_idurl is not None and prev_position is not None:
                    if prev_broker_idurl != new_broker_idurl or prev_position != new_position:
                        if _Debug:
                            lg.out(_DebugLevel, '        new json data have same revision but different broker info, store operation FAILED')
                        raise ValueError('new json data have same revision but different broker info, current revision is %d ' % prev_revision)
    if _Debug:
        lg.out(_DebugLevel, '        new json data is valid and matching existing DHT record, store OK')
    return True


def validate_before_request(key, **kwargs):
    """
    Will be executed on receiver side for each "direct" key request on DHT
    """
    if _Debug:
        lg.out(_DebugLevel, 'dht_service.validate_before_request key=[%s]' % key)
    return True


def validate_data_written(store_results, key, json_data, result_defer):
    """
    Will be executed on sender side for each (key,value) set request on DHT
    """
    nodes = store_results
    results_collected = False
    if isinstance(store_results, tuple):
        results_collected = True
        nodes = store_results[0]
    if _Debug:
        lg.out(_DebugLevel, 'dht_service.validate_data_written key=[%s]  %s  collected=%s  nodes=%r' % (key, type(store_results), results_collected, nodes))
    if results_collected:
        for result in store_results[1]:
            try:
                success = strng.to_text(result[0])
                response = result[1]
                success_str = repr(success)
                response_str = repr(response)
            except Exception as exc:
                lg.exc()
                result_defer.errback(exc)
                return None
            if success_str.count('TimeoutError') or response_str.count('TimeoutError'):
                continue
            if not success:
                if _Debug:
                    lg.out(_DebugLevel, '    store operation failed: %r' % response)
                result_defer.errback(ValueError(response))
                return None
            if response != 'OK':
                if _Debug:
                    lg.out(_DebugLevel, '    store operation failed, unexpected response received: %r' % response)
                result_defer.errback(ValueError(response))
                return None
        result_defer.callback(nodes)
        return None
    if isinstance(store_results, list):
        result_defer.callback(nodes)
    else:
        result_defer.errback(Exception('store operation failed, result is %s' % type(store_results)))
    return None


def validate_before_send(value, key, rules, populate_meta_fields):
    """
    Will be executed on sender side before every (key,value) set request towards DHT
    """
    if _Debug:
        lg.out(_DebugLevel, 'dht_service.validate_before_send for key [%s]' % key)
    if populate_meta_fields:
        if value.get('key') != key:
            value['key'] = key
        if value.get('type') != rules['type'][0]['arg']:
            value['type'] = rules['type'][0]['arg']
    return validate_rules(value, key, rules, raise_for_result=False, populate_meta_fields=populate_meta_fields)


def validate_after_receive(value, key, rules, result_defer, raise_for_result, populate_meta_fields):
    """
    Will be executed on sender side for each (key,value) get response from DHT
    """
    if _Debug:
        lg.out(_DebugLevel, 'dht_service.validate_after_receive for key [%s]' % key)
    response = validate_rules(value, key, rules, result_defer=result_defer, raise_for_result=raise_for_result, populate_meta_fields=populate_meta_fields)
    if not isinstance(response, dict):
        if populate_meta_fields:
            ret = {
                'key': key,
                'type': rules['type'][0]['arg'],
                'closest': response,
            }
            return ret
    if populate_meta_fields:
        if response.get('key') != key:
            response['key'] = key
        if response.get('type') != rules['type'][0]['arg']:
            response['type'] = rules['type'][0]['arg']
    return response


#------------------------------------------------------------------------------


def get_valid_data(key, rules={}, raise_for_result=False, return_details=False, layer_id=0, use_cache_ttl=None, update_cache=True):
    ret = Deferred()
    if use_cache_ttl is not None:
        d = get_cached_json_value(key, layer_id=layer_id, cache_ttl=use_cache_ttl)
    else:
        d = get_json_value(key, layer_id=layer_id, update_cache=update_cache)
    d.addCallback(validate_after_receive, key=key, rules=rules, result_defer=ret, raise_for_result=raise_for_result, populate_meta_fields=return_details)
    d.addErrback(ret.errback)
    return ret


def set_valid_data(key, json_data, age=0, expire=KEY_EXPIRE_MAX_SECONDS, rules={}, collect_results=False, layer_id=0):
    valid_json_data = validate_before_send(value=json_data, key=key, rules=rules, populate_meta_fields=True)
    if valid_json_data is None or not isinstance(valid_json_data, dict):
        return fail(Exception('invalid data going to be written, validation failed'))
    ret = Deferred()
    d = set_json_value(key, json_data=json_data, age=age, expire=expire, collect_results=collect_results, layer_id=layer_id)
    d.addCallback(validate_data_written, key, json_data, ret)
    d.addErrback(ret.errback)
    return ret


#------------------------------------------------------------------------------


def write_verify_republish_data(key, json_data, age=0, expire=KEY_EXPIRE_MAX_SECONDS, rules={}, layer_id=0):
    try:
        raw_value = jsn.dumps(json_data, indent=0, sort_keys=True, separators=(',', ':'))
    except:
        return fail(Exception('bad input json data'))

    ret = Deferred()
    _found_nodes = None
    _write_response = None
    _join = Deferred()
    _join.addCallback(_do_verify)
    _join.addErrback(lg.errback, debug=_Debug, debug_level=_DebugLevel, method='write_verify_republish_data')

    def _some_nodes_found(nodes):
        global _write_response
        global _found_nodes
        global _join
        if _Debug:
            lg.out(_DebugLevel, 'dht_service._some_nodes_found : %r' % nodes)
        if len(nodes) > 0:
            _found_nodes = nodes
        else:
            _found_nodes = []
        if _write_response:
            _join.callback(_write_response, _found_nodes)
        return nodes

    def _nodes_not_found(err):
        global _found_nodes
        global _join
        if _Debug:
            lg.out(_DebugLevel, 'dht_service._nodes_not_found err=%s' % str(err))
        _found_nodes = []
        _join.cancel()
        del _join
        ret.errback(err)
        return err

    def _write_ok(write_result):
        global _write_response
        global _found_nodes
        global _join
        if _Debug:
            lg.out(_DebugLevel, 'dht_service._write_ok : %r' % write_result)
        _write_response = write_result
        if _found_nodes is not None:
            _join.callback(_write_response, _found_nodes)
        return write_result

    def _write_failed(err):
        global _join
        if _Debug:
            lg.out(_DebugLevel, 'dht_service._write_failed  err=%r' % err)
        _join.cancel()
        del _join
        ret.errback(err)
        return err

    def _do_verify(write_response, found_nodes):
        if _Debug:
            lg.out(_DebugLevel, 'dht_service._do_verify  %r via nodes: %r' % (write_response, found_nodes))
        for node in found_nodes:
            node.request(b'verify_update', key, raw_value, age, expire)
        ret.callback(write_response, found_nodes)
        return None

    new_key = random_key()
    d_observer = find_node(new_key, layer_id=layer_id)
    d_observer.addCallback(_some_nodes_found)
    d_observer.addErrback(_nodes_not_found)

    d_write = set_valid_data(key=key, json_data=json_data, age=age, expire=expire, rules=rules, layer_id=layer_id)
    d_write.addCallback(_write_ok)
    d_write.addErrback(_write_failed)

    return ret


#------------------------------------------------------------------------------


def on_nodes_found(result, node_id64):
    global _ActiveLookup
    global _ActiveLookupLayerID
    on_success(result, 'find_node', node_id64)
    if _Debug:
        lg.out(_DebugLevel, 'dht_service.on_nodes_found   node_id=[%s], %d nodes found' % (node_id64, len(result)))
    # _ActiveLookupLayerID = None
    # _ActiveLookup = None
    return result


def on_lookup_failed(result, node_id64):
    global _ActiveLookup
    global _ActiveLookupLayerID
    on_error(result, 'find_node', node_id64)
    if _Debug:
        lg.out(_DebugLevel, 'dht_service.on_lookup_failed   node_id=[%s], result=%s' % (node_id64, result))
    # _ActiveLookupLayerID = None
    # _ActiveLookup = None
    return result


def find_node(node_id, layer_id=0):
    global _ActiveLookup
    global _ActiveLookupLayerID
    if _ActiveLookup and not _ActiveLookup.called and _ActiveLookupLayerID == layer_id:
        if _Debug:
            lg.out(_DebugLevel, 'dht_service.find_node SKIP, already started')
        return _ActiveLookup
    count('find_node')
    if not node():
        return fail(Exception('DHT service is off'))
    if layer_id not in node().active_layers:
        return fail(Exception('DHT layer %d is not active' % layer_id))
    node_id64 = node_id
    if _Debug:
        lg.out(_DebugLevel, 'dht_service.find_node   node_id=[%s]  layer_id=%d' % (node_id64, layer_id))
    _ActiveLookupLayerID = layer_id
    _ActiveLookup = node().iterativeFindNode(node_id, layerID=layer_id)
    _ActiveLookup.addErrback(on_lookup_failed, node_id64)
    _ActiveLookup.addCallback(on_nodes_found, node_id64)
    return _ActiveLookup


#------------------------------------------------------------------------------


def get_node_data(key, layer_id=0):
    if not node():
        if _Debug:
            lg.out(_DebugLevel, 'dht_service.get_node_data local node is not ready')
        return None
    count('get_node_data')
    if layer_id not in node().data:
        if _Debug:
            lg.out(_DebugLevel, 'dht_service.get_node_data   layer_id=%d   is not exist' % layer_id)
        return None
    key = strng.to_text(key)
    if key not in node().data[layer_id]:
        if _Debug:
            lg.out(_DebugLevel, 'dht_service.get_node_data   key=[%s] not exist   layer_id=%d' % (key, layer_id))
        return None
    value = node().data[layer_id][key]
    if _Debug:
        lg.out(_DebugLevel, 'dht_service.get_node_data key=[%s] read %d bytes   layer_id=%d' % (key, len(strng.to_text(value) or ''), layer_id))
    return value


def set_node_data(key, value, layer_id=0):
    if not node():
        lg.warn('DHT node is not ready yet, not able to store key %r for layer_id=%d locally' % (key, layer_id))
        return False
    count('set_node_data')
    if layer_id not in node().data:
        node().data[layer_id] = {}
    key = strng.to_text(key)
    node().data[layer_id][key] = value
    if _Debug:
        lg.out(_DebugLevel, 'dht_service.set_node_data key=[%s] wrote %d bytes  layer_id=%d' % (key, len(strng.to_text(value) or ''), layer_id))
    return True


def delete_node_data(key, layer_id=0):
    if not node():
        if _Debug:
            lg.out(_DebugLevel, 'dht_service.delete_node_data local node is not ready')
        return False
    count('delete_node_data')
    if layer_id not in node().data:
        if _Debug:
            lg.out(_DebugLevel, 'dht_service.delete_node_data layer_id=%d  is not exist' % layer_id)
        return False
    key = strng.to_text(key)
    if key not in node().data[layer_id]:
        if _Debug:
            lg.out(_DebugLevel, 'dht_service.delete_node_data key=[%s] not exist' % key)
        return False
    node().data[layer_id].pop(key)
    if _Debug:
        lg.out(_DebugLevel, 'dht_service.delete_node_data key=[%s]   layer_id=%d' % (key, layer_id))
    return True


def dump_local_db(value_as_json=False):
    if not node():
        if _Debug:
            lg.out(_DebugLevel, 'dht_service.dump_local_db local node is not ready')
        return None
    result = {}
    try:
        for layerID in node()._dataStores.keys():
            l = []
            for itm in node()._dataStores[layerID].getAllItems():
                if value_as_json:
                    if isinstance(itm['value'], dict):
                        _j = jsn.dumps(itm['value'], keys_to_text=True, errors='ignore')
                        itm['value'] = jsn.loads_text(_j, errors='ignore')
                    else:
                        try:
                            itm['value'] = jsn.loads_text(itm['value'], errors='ignore')
                        except:
                            itm['value'] = strng.to_text(itm['value'])
                itm['scope'] = 'global'
                l.append(itm)
            for k, v in node().data[layerID].items():
                l.append({
                    'key': k,
                    'value': v,
                    'scope': 'node',
                })
            result[layerID] = l
    except:
        lg.exc()
    return result


#------------------------------------------------------------------------------


def cache():
    global _Cache
    return _Cache


def load_cache(cache_dir_path):
    global _Cache
    _Cache.clear()
    total_records = 0
    records_per_layer = {}
    for layer_id_str in os.listdir(cache_dir_path):
        layer_id = int(layer_id_str)
        if layer_id not in _Cache:
            _Cache[layer_id] = {}
        records_per_layer[layer_id] = 0
        for hash_key in os.listdir(cache_dir_path):
            _Cache[layer_id][hash_key] = jsn.loads_text(local_fs.ReadTextFile(os.path.join(cache_dir_path, hash_key)))
            total_records += 1
            records_per_layer[layer_id] += 1
    if _Debug:
        lg.args(_DebugLevel, total_records=total_records, records_per_layer=records_per_layer)
    return total_records


def store_cached_key(hash_key, json_value, layer_id=0, timestamp=None):
    global _Cache
    if not timestamp:
        timestamp = utime.utcnow_to_sec1970()
    if layer_id not in _Cache:
        _Cache[layer_id] = {}
    cached_json_record = {
        'v': json_value,
        't': timestamp,
    }
    dht_dir_path = settings.ServiceDir('service_entangled_dht')
    layer_cache_dir_path = os.path.join(dht_dir_path, 'cache', strng.to_text(layer_id))
    if not os.path.isdir(layer_cache_dir_path):
        os.makedirs(layer_cache_dir_path)
    cached_record_file_path = os.path.join(layer_cache_dir_path, hash_key)
    if not local_fs.WriteTextFile(cached_record_file_path, jsn.dumps(cached_json_record)):
        lg.err('failed to store cached dht key %r in layer %d' % (hash_key, layer_id))
        return False
    _Cache[layer_id][hash_key] = cached_json_record
    if _Debug:
        lg.args(_DebugLevel, hash_key=hash_key, layer_id=layer_id, timestamp=timestamp, cached_records=len(_Cache[layer_id]))
    return True


def get_cached_value(hash_key, layer_id=0):
    global _Cache
    value = _Cache.get(layer_id, {}).get(hash_key)
    if _Debug:
        lg.args(_DebugLevel, layer_id=layer_id, hash_key=hash_key, value_exist=(value is not None))
    return value


def on_json_response_to_be_cached(json_value, key, layer_id):
    if not json_value:
        return json_value
    hash_key = key_to_hash(key)
    store_cached_key(hash_key, json_value, layer_id)
    return json_value


def get_cached_json_value(key, layer_id=0, cache_ttl=DEFAULT_CACHE_TTL):
    hash_key = key_to_hash(key)
    cached_record = get_cached_value(hash_key)
    if not cached_record:
        return get_json_value(key, layer_id=layer_id, update_cache=True)
    if utime.utcnow_to_sec1970() - int(cached_record['t']) > cache_ttl:
        return get_json_value(key, layer_id=layer_id, update_cache=True)
    if _Debug:
        lg.out(_DebugLevel, 'dht_service.get_cached_json_value key=[%r] layer_id=%d cache_ttl=%d' % (key, layer_id, cache_ttl))
    ret = Deferred()
    ret.callback(cached_record['v'])
    return ret


#------------------------------------------------------------------------------


class DHTNode(MultiLayerNode):
    def __init__(self, udpPort=4000, dataStores=None, routingTables=None, networkProtocol=None, nodeID=None):
        super(DHTNode, self).__init__(
            udpPort=udpPort,
            dataStores=dataStores,
            routingTables=routingTables,
            networkProtocol=networkProtocol,
            id=nodeID,
        )
        self._counter = count
        self.data = {
            0: {},
        }
        if dataStores:
            for layer_id in dataStores.keys():
                self.data[layer_id] = {}
        self.expire_task = LoopingCall(self.expire)
        self.rpc_callbacks = {}

    def add_rpc_callback(self, rpc_method_name, cb):
        self.rpc_callbacks[rpc_method_name] = cb

    def remove_rpc_callback(self, rpc_method_name):
        self.rpc_callbacks.pop(rpc_method_name, None)

    def reset_my_dht_id(self, new_id=None):
        self._routingTable = TreeRoutingTable(self.id)
        h = hashlib.sha1()
        h.update(b'nodeState')
        nodeStateKey = h.hexdigest()
        if nodeStateKey in self._dataStore:
            json_state = self._dataStore[nodeStateKey]
            state = json.loads(json_state)
            self.id = state['id']
            self._routingTable = TreeRoutingTable(self.id)
            for contactTriple in state['closestNodes']:
                contact = Contact(encoding.to_text(contactTriple[0]), contactTriple[1], contactTriple[2], self._protocol)
                self._routingTable.addContact(contact)
            if _Debug:
                print('[DHT NODE]    found "nodeState" key in local db and added %d contacts to routing table' % len(state['closestNodes']))
        else:
            self.id = self._generateID()
            self._routingTable = TreeRoutingTable(self.id)
        self._counter = None

    def expire(self):
        now = utime.utcnow_to_sec1970()
        for layer_id in self._dataStores.keys():
            expired_keys = []
            for key in self._dataStores[layer_id].keys():
                if key == self.nodeStateKey:
                    continue
                item_data = self._dataStores[layer_id].getItem(key)
                if item_data:
                    originaly_published = item_data.get('originallyPublished')
                    expireSeconds = item_data.get('expireSeconds')
                    if expireSeconds and originaly_published:
                        age = now - originaly_published
                        if age > expireSeconds:
                            expired_keys.append(key)
            for key in expired_keys:
                if _Debug:
                    lg.out(_DebugLevel, 'dht_service.expire   [%s] removed from layer %d' % (key, layer_id))
                del self._dataStores[layer_id][key]

    @rpcmethod
    def store(self, key, value, originalPublisherID=None, age=0, expireSeconds=KEY_EXPIRE_MAX_SECONDS, **kwargs):
        count('store_dht_service')
        layerID = kwargs.pop('layerID', 0)
        if _Debug:
            lg.out(_DebugLevel, 'dht_service.DHTNode.store key=[%s] for %d seconds layerID=%d' % (key, expireSeconds, layerID))

        if 'store' in self.rpc_callbacks and layerID in self.active_layers:
            # TODO: add signature validation to be sure this is the owner of that key:value pair
            self.rpc_callbacks['store'](key=key, value=value, originalPublisherID=originalPublisherID, age=age, expireSeconds=expireSeconds, layerID=layerID, **kwargs)

        return super(DHTNode, self).store(key=key, value=value, originalPublisherID=originalPublisherID, age=age, expireSeconds=expireSeconds, layerID=layerID, **kwargs)

    @rpcmethod
    def request(self, key, **kwargs):
        count('request')
        layerID = kwargs.get('layerID', 0)
        if layerID not in self.active_layers:
            if _Debug:
                lg.out(_DebugLevel, 'dht_service.DHTNode.request key=[%r] SKIP because layer %d is not active' % (key, layerID))
            return {
                key: 0,
                layerID: layerID,
            }
        if _Debug:
            lg.out(_DebugLevel, 'dht_service.DHTNode.request key=[%r] layerID=%d' % (key, layerID))
        if 'request' in self.rpc_callbacks:
            self.rpc_callbacks['request'](key=key, layerID=layerID)
        value = get_node_data(key, layer_id=layerID)
        if value is None:
            value = 0
        if _Debug:
            lg.out(_DebugLevel, '    read internal value: %r' % value)
        return {
            key: value,
            layerID: layerID,
        }

    @rpcmethod
    def verify_update(self, key, value, originalPublisherID=None, age=0, expireSeconds=KEY_EXPIRE_MAX_SECONDS, **kwargs):
        count('verify_update')
        if _Debug:
            lg.out(_DebugLevel, 'dht_service.DHTNode.verify_update key=[%s]' % strng.to_text(key, errors='ignore')[:10])
        return True

    @rpcmethod
    def findNode(self, key, **kwargs):
        count('findNode')
        layerID = kwargs.get('layerID', 0)
        if _Debug:
            lg.out(_DebugLevel, 'dht_service.DHTNode.findNode key=[%s] layerID=%d' % (key, layerID))
        return super(DHTNode, self).findNode(key, **kwargs)

    @rpcmethod
    def findValue(self, key, **kwargs):
        count('findValue')
        layerID = kwargs.get('layerID', 0)
        if _Debug:
            lg.out(_DebugLevel, 'dht_service.DHTNode.findValue key=[%s] layerID=%d' % (key, layerID))
        return super(DHTNode, self).findValue(key, **kwargs)

    @rpcmethod
    def ping(self):
        count('ping')
        if _Debug:
            lg.out(_DebugLevel, 'dht_service.DHTNode.ping')
        return super(DHTNode, self).ping()

    def reconnect(self, knownNodeAddresses=None):
        """
        TODO: need to restart _scheduleNextNodeRefresh.
        """
        d = Deferred()
        if not self.listener:
            d.errback(Exception('Listener is not started yet'))
            return d
        d.callback(1)
        return d


#------------------------------------------------------------------------------


class DHTProtocol(KademliaMultiLayerProtocol):
    def __init__(self, node, msgEncoder=encoding.Bencode(), msgTranslator=msgformat.MultiLayerFormat()):
        KademliaMultiLayerProtocol.__init__(self, node, msgEncoder=msgEncoder, msgTranslator=msgTranslator)
        self._counter = count


#         self.receiving_queue = []
#         self.receiving_worker = None
#         self.sending_queue = []
#         self.sending_worker = None

#     def datagramReceived(self, datagram, address):
#         count('dht_datagramReceived')
#         if len(self.receiving_queue) > RECEIVING_QUEUE_LENGTH_CRITICAL:
#             lg.warn('incoming DHT traffic too high, items to process: %d' % len(self.receiving_queue))
#         self.receiving_queue.append((datagram, address, ))
#         if self.receiving_worker is None:
#             self._process_incoming()  # self.receiving_worker = reactor.callLater(0, self._process_incoming)

#     def _process_incoming(self):
#         if len(self.receiving_queue) == 0:
#             self.receiving_worker = None
#             return
#         datagram, address = self.receiving_queue.pop(0)
#         KademliaMultiLayerProtocol.datagramReceived(self, datagram, address)
#         t = 0
#         if len(self.receiving_queue) > RECEIVING_QUEUE_LENGTH_CRITICAL / 2:
#             t = RECEIVING_FREQUENCY_SEC
#         self.receiving_worker = reactor.callLater(t, self._process_incoming)  #@UndefinedVariable

#     def _send(self, data, rpcID, address):
#         count('dht_send')
#         if _Debug:
#             if len(self.sending_queue) > 50:
#                 lg.warn('outgoing DHT traffic too high, items to send: %d' % len(self.sending_queue))
#         self.sending_queue.append((data, rpcID, address, ))
#         if self.receiving_worker is None:
#             self._process_outgoing() # self.receiving_worker = reactor.callLater(0, self._process_outgoing)

#     def _process_outgoing(self):
#         if len(self.sending_queue) == 0:
#             self.sending_worker = None
#             return
#         data, rpcID, address = self.sending_queue.pop(0)
#         KademliaMultiLayerProtocol._send(self, data, rpcID, address)
#         t = 0
#         if len(self.sending_queue) > SENDING_QUEUE_LENGTH_CRITICAL:
#             t = SENDING_FREQUENCY_SEC
#         self.sending_worker = reactor.callLater(t, self._process_outgoing)  #@UndefinedVariable

#------------------------------------------------------------------------------


def parseCommandLine():
    oparser = optparse.OptionParser()
    oparser.add_option('-p', '--udpport', dest='udpport', type='int', help='specify UDP port for DHT network')
    oparser.set_default('udpport', settings.DefaultDHTPort())
    oparser.add_option('-d', '--dhtdb', dest='dhtdb', help='specify DHT folder location')
    oparser.set_default('dhtdb', settings.ServiceDir('service_entangled_dht'))
    oparser.add_option('-s', '--seeds', dest='seeds', help='specify list of DHT seed nodes')
    oparser.set_default('seeds', '')
    oparser.add_option('-l', '--layers', dest='layers', help='specify list of layers to be created')
    oparser.set_default('layers', '')
    oparser.add_option('-a', '--attach_layer', dest='attach_layer', help='specify which layer to be attached at start')
    oparser.set_default('attach_layer', '')
    oparser.add_option('-j', '--join_layer', dest='join_layer', help='specify which layer to be joined at start')
    oparser.set_default('join_layer', '')
    oparser.add_option('-w', '--wait', dest='delayed', type='int', help='wait N seconds before join the network')
    oparser.set_default('delayed', 0)
    (options, args) = oparser.parse_args()
    return options, args


def main(options=None, args=None):
    from bitdust.dht import dht_relations

    if options is None and args is None:
        (options, args) = parseCommandLine()

    else:
        (_options, _args) = parseCommandLine()
        if options is None:
            options = _options
        if args is None:
            args = _args

    if not os.path.exists(options.dhtdb):
        os.makedirs(options.dhtdb)

    connect_layers = [int(l) for l in options.layers.split(',') if l]
    init(udp_port=options.udpport, dht_dir_path=options.dhtdb, open_layers=connect_layers)
    lg.out(_DebugLevel, 'Init   udpport=%d   dhtdb=%s   node=%r   connect_layers=%r' % (options.udpport, options.dhtdb, node(), connect_layers))

    seeds = []

    def _go():
        lg.out(_DebugLevel, 'DHT node is active, layers: %r' % node().layers)

        try:
            if len(args) == 0:
                pass

            elif len(args) > 0:

                cmd = args[0]

                lg.info('COMMAND: %r' % cmd)

                def _r(x):
                    lg.info('RESULT "%s": %r' % (cmd, x))
                    reactor.stop()  #@UndefinedVariable

                if cmd == 'get':
                    get_value(args[1], layer_id=0 if len(args) < 3 else int(args[2])).addBoth(_r)
                elif cmd == 'set':
                    set_value(args[1], args[2], expire=int(args[3]), layer_id=0 if len(args) < 5 else int(args[4])).addBoth(_r)
                elif cmd == 'get_json':
                    get_json_value(args[1], layer_id=0 if len(args) < 3 else int(args[2])).addBoth(_r)
                elif cmd == 'set_json':
                    set_json_value(args[1], jsn.loads(args[2]), expire=int(args[3]), layer_id=0 if len(args) < 3 else int(args[2])).addBoth(_r)
                elif cmd == 'get_valid_data':
                    get_valid_data(args[1], rules=jsn.loads(args[2]), return_details=True, layer_id=0 if len(args) < 4 else int(args[3])).addBoth(_r)
                elif cmd == 'set_valid_data':
                    set_valid_data(args[1], jsn.loads(args[2]), expire=int(args[3]), rules=jsn.loads(args[4]), layer_id=0 if len(args) < 6 else int(args[5])).addBoth(_r)
                elif cmd == 'read_customer_suppliers':
                    dht_relations.read_customer_suppliers(args[1]).addBoth(_r)
                elif cmd == 'write_customer_suppliers':
                    dht_relations.write_customer_suppliers(args[1], args[2].split(',')).addBoth(_r)
                elif cmd == 'write_verify_republish':
                    write_verify_republish_data(args[1], args[2], expire=int(args[3])).addBoth(_r)
                elif cmd == 'find':
                    find_node(key_to_hash(args[1])).addBoth(_r)
                elif cmd == 'ping':
                    find_node(random_key()).addBoth(_r)
                elif cmd == 'get_node_data':
                    pprint.pprint(get_node_data(args[1]))
                elif cmd == 'observe_data':

                    def _p(val, n):
                        print('observed', n, val)

                    def _o(result):
                        for n in result:
                            d = n.request(args[2])
                            d.addCallback(_p, n)

                    d = find_node(key_to_hash(args[1]))
                    d.addErrback(_r)
                    d.addCallback(_o)
                elif cmd == 'discover':

                    def _l(x):
                        lg.info(x)
                        find_node(random_key()).addBoth(_l)

                    _l('')
                elif cmd == 'dump_db':
                    pprint.pprint(dump_local_db(value_as_json=True))
                elif cmd == 'create_layer':
                    open_layer(
                        layer_id=int(args[1]),
                        seed_nodes=[(s[0], int(s[1])) for s in args[2].split(',') if s],
                        dht_dir_path=options.dhtdb,
                        connect_now=True,
                    ).addBoth(_r)
        except:
            lg.exc()

    possible_seeds = options.seeds
    if possible_seeds in [
        'genesis',
        'root',
        b'genesis',
        b'root',
    ]:
        # "genesis" node must not connect anywhere
        possible_seeds = []

    lg.out(_DebugLevel, 'options.seeds=%r' % options.seeds)
    lg.out(_DebugLevel, 'possible_seeds=%r' % possible_seeds)

    if options.seeds in [
        'genesis',
        'root',
        b'genesis',
        b'root',
    ]:
        lg.out(_DebugLevel, 'Starting genesis node!!!!!!!!!!!!!!!!!!!!')

    else:
        for dht_node_str in options.seeds.split(','):
            if dht_node_str.strip():
                try:
                    dht_node = dht_node_str.strip().split(':')
                    dht_node_host = dht_node[0].strip()
                    dht_node_port = int(dht_node[1].strip())
                except:
                    continue
                seeds.append((
                    dht_node_host,
                    dht_node_port,
                ))

        if not seeds:
            seeds = known_nodes.nodes()

        lg.out(_DebugLevel, 'Seed nodes: %s' % seeds)

    if options.delayed:
        lg.out(_DebugLevel, 'Wait %d seconds before join the network' % options.delayed)
        import time
        time.sleep(options.delayed)

    def _layers_connected(allresults):
        lg.out(_DebugLevel, 'Layers are connected: %r' % allresults)
        _go()

    def _connected(nodes):
        lg.out(_DebugLevel, 'Connected, known contacts: %r' % nodes)
        l = []
        if options.attach_layer:
            layer_id = int(options.attach_layer)
            l.append(connect(seeds, layer_id=layer_id, attach=True))
        if options.join_layer:
            layer_id = int(options.join_layer)
            l.append(connect(seeds, layer_id=layer_id, attach=False))
        d = DeferredList(l)
        d.addBoth(_layers_connected)

    def _start():
        connect(seeds).addBoth(_connected)

    reactor.callWhenRunning(_start)  # @UndefinedVariable
    reactor.run()  #@UndefinedVariable


#------------------------------------------------------------------------------

if __name__ == '__main__':
    from bitdust.dht import dht_service
    bpio.init()
    settings.init()
    lg.set_debug_level(20)
    dht_service.main()
    settings.shutdown()



#!/usr/bin/python
# proxy_interface.py
#
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (proxy_interface.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: proxy_interface.

This is a client side part of the PROXY transport plug-in.
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
import six

#------------------------------------------------------------------------------

_Debug = False

#------------------------------------------------------------------------------

from twisted.web import xmlrpc
from twisted.internet.defer import succeed, fail, Deferred

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.main import settings

from bitdust.contacts import identitycache

#------------------------------------------------------------------------------

_GateProxy = None

#------------------------------------------------------------------------------


def proxy():
    global _GateProxy
    return _GateProxy


#------------------------------------------------------------------------------


class GateInterface():
    def init(self, xml_rpc_url_or_object):
        """
        """
        global _GateProxy
        if _Debug:
            lg.out(4, 'proxy_interface.init')
        from bitdust.transport.proxy import proxy_receiver
        from bitdust.transport.proxy import proxy_sender
        if isinstance(xml_rpc_url_or_object, six.string_types):
            _GateProxy = xmlrpc.Proxy(xml_rpc_url_or_object, allowNone=True)
        else:
            _GateProxy = xml_rpc_url_or_object
        proxy_receiver.A('init')
        proxy_sender.A('init')
        proxy().callRemote('transport_initialized', 'proxy')
        return True

    def shutdown(self):
        """
        """
        if _Debug:
            lg.out(4, 'proxy_interface.shutdown')
        from bitdust.transport.proxy import proxy_receiver
        from bitdust.transport.proxy import proxy_sender
        ret = self.disconnect()
        proxy_receiver.A('shutdown')
        proxy_sender.A('shutdown')
        global _GateProxy
        if _GateProxy:
            _GateProxy = None
        return ret

    def connect(self, options):
        """
        """
        if _Debug:
            lg.out(4, 'proxy_interface.connect %s' % str(options))
        from bitdust.transport.proxy import proxy_receiver
        from bitdust.transport.proxy import proxy_sender
        if settings.enablePROXYreceiving():
            proxy_receiver.A('start', options)
        else:
            lg.warn('proxy transport receiving is disabled')
            interface_receiving_failed()
            return False
        if settings.enablePROXYsending():
            proxy_sender.A('start', options)
        return succeed(True)

    def disconnect(self):
        """
        """
        if _Debug:
            lg.out(4, 'proxy_interface.disconnect')
        from bitdust.transport.proxy import proxy_receiver
        from bitdust.transport.proxy import proxy_sender
        if not proxy_receiver.A():
            lg.warn('proxy_receiver is not ready')
            interface_disconnected()
        elif proxy_receiver.A().state != 'LISTEN':
            lg.warn('proxy_receiver is not listening now')
            interface_disconnected()
        else:
            proxy_receiver.A('stop')
        proxy_sender.A('stop')
        return succeed(True)

    def build_contacts(self, id_obj):
        """
        """
        from bitdust.transport.proxy import proxy_receiver
        # from bitdust.userid import my_id
        # current_identity_contacts = my_id.getLocalIdentity().getContacts()
        if not proxy_receiver.GetRouterIdentity():
            # if not yet found one node to route your traffic - do nothing
            if _Debug:
                lg.out(4, 'proxy_interface.build_contacts SKIP, router not yet found')
            return []  # current_identity_contacts
        if not proxy_receiver.ReadMyOriginalIdentitySource():
            # if we did not save our original identity we will have troubles contacting remote node
            if _Debug:
                lg.out(4, 'proxy_interface.build_contacts SKIP, original identity was not saved')
            return []  # current_identity_contacts
        # switch contacts - use router contacts instead of my
        # he will receive all packets addressed to me and redirect to me
        result = proxy_receiver.GetRouterIdentity().getContacts()
        if _Debug:
            lg.out(4, 'proxy_interface.build_contacts %s : %r' % (proxy_receiver.GetRouterIdentity().getIDName(), result))
        return result

    def verify_contacts(self, id_obj):
        """
        Check if router is ready and his contacts exists in that identity.
        """
        from bitdust.transport.proxy import proxy_receiver
        if not proxy_receiver.A() or not proxy_receiver.GetRouterIDURL() or not proxy_receiver.GetRouterIdentity():
            # if not yet found any node to route your traffic - do nothing
            if _Debug:
                lg.out(4, 'proxy_interface.verify_contacts returning True : router not yet found')
            return True
        if not proxy_receiver.ReadMyOriginalIdentitySource():
            if _Debug:
                lg.out(4, 'proxy_interface.verify_contacts returning False : my original identity is empty')
            return False
        result = Deferred()

        def _finish_verification(res):
            if _Debug:
                lg.out(4, 'proxy_interface._finish_verification')
            try:
                cached_id = identitycache.FromCache(proxy_receiver.GetRouterIDURL())
                if not cached_id:
                    if _Debug:
                        lg.out(4, '    returning False: router identity is not cached')
                    res.callback(False)
                    return False
                if not proxy_receiver.GetRouterIdentity():
                    if _Debug:
                        lg.out(4, '    returning False : router identity is None or router is not ready yet')
                    return True
                if cached_id.serialize() != proxy_receiver.GetRouterIdentity().serialize():
                    if _Debug:
                        lg.out(4, 'proxy_interface.verify_contacts return False: cached copy is different')
                        lg.out(20, '\n%s\n' % cached_id.serialize(as_text=True))
                        lg.out(20, '\n%s\n' % proxy_receiver.GetRouterIdentity().serialize(as_text=True))
                    res.callback(False)
                    return
                router_contacts = proxy_receiver.GetRouterIdentity().getContactsByProto()
                if len(router_contacts) != id_obj.getContactsNumber():
                    if _Debug:
                        lg.out(4, '    returning False: router contacts is different')
                    res.callback(False)
                    return False
                for proto, contact in id_obj.getContactsByProto().items():
                    if proto not in list(router_contacts.keys()):
                        if _Debug:
                            lg.out(4, '    returning False: [%s] is not present in router contacts' % proto)
                        res.callback(False)
                        return False
                    if router_contacts[proto] != contact:
                        if _Debug:
                            lg.out(4, '    returning False: [%s] contact is different in router id' % proto)
                        res.callback(False)
                        return False
                if _Debug:
                    lg.out(4, '    returning True : my contacts and router contacts is same')
                res.callback(True)
                return True
            except:
                lg.exc()
                res.callback(True)
                return True

        d = identitycache.immediatelyCaching(proxy_receiver.GetRouterIDURL())
        d.addCallback(lambda src: _finish_verification(result))
        d.addErrback(lambda err: result.callback(False) and None)
        return result

    def list_sessions(self):
        """
        """
        from bitdust.transport.proxy import proxy_receiver
        from bitdust.transport.proxy import proxy_sender
        result = []
        if proxy_receiver.A():
            result.append(proxy_receiver.A())
        if proxy_sender.A():
            result.append(proxy_sender.A())
        return result

    def list_streams(self, sorted_by_time=True):
        """
        """
        return []


#------------------------------------------------------------------------------


def proxy_errback(x):
    if _Debug:
        lg.out(6, 'proxy_interface.proxy_errback ERROR %s' % x)
    return None


#------------------------------------------------------------------------------


def interface_transport_initialized(xmlrpcurl):
    if proxy():
        return proxy().callRemote('transport_initialized', 'proxy', xmlrpcurl).addErrback(proxy_errback)
    lg.warn('transport_proxy is not ready')
    return fail(Exception('transport_proxy is not ready')).addErrback(proxy_errback)


def interface_receiving_started(host, new_options={}):
    if proxy():
        return proxy().callRemote('receiving_started', 'proxy', host, new_options).addErrback(proxy_errback)
    lg.warn('transport_proxy is not ready')
    return fail(Exception('transport_proxy is not ready')).addErrback(proxy_errback)


def interface_receiving_failed(error_code=None):
    if proxy():
        return proxy().callRemote('receiving_failed', 'proxy', error_code).addErrback(proxy_errback)
    lg.warn('transport_proxy is not ready')
    return fail(Exception('transport_proxy is not ready')).addErrback(proxy_errback)


def interface_disconnected(result=None):
    if proxy():
        return proxy().callRemote('disconnected', 'proxy', result).addErrback(proxy_errback)
    lg.warn('transport_proxy is not ready')
    # return fail(Exception('transport_proxy is not ready')).addErrback(proxy_errback)
    return succeed(result)


def interface_register_file_sending(host, receiver_idurl, filename, size=0, description=''):
    if proxy():
        return proxy().callRemote(
            'register_file_sending',
            'proxy',
            host,
            receiver_idurl,
            filename,
            size,
            description,
        ).addErrback(proxy_errback)
    lg.warn('transport_proxy is not ready')
    return fail(Exception('transport_proxy is not ready')).addErrback(proxy_errback)


def interface_register_file_receiving(host, sender_idurl, filename, size=0):
    if proxy():
        return proxy().callRemote(
            'register_file_receiving',
            'proxy',
            host,
            sender_idurl,
            filename,
            size,
        ).addErrback(proxy_errback)
    lg.warn('transport_proxy is not ready')
    return fail(Exception('transport_proxy is not ready')).addErrback(proxy_errback)


def interface_unregister_file_sending(transfer_id, status, size=0, error_message=None):
    if proxy():
        return proxy().callRemote(
            'unregister_file_sending',
            transfer_id,
            status,
            size,
            error_message,
        ).addErrback(proxy_errback)
    lg.warn('transport_proxy is not ready')
    return fail(Exception('transport_proxy is not ready')).addErrback(proxy_errback)


def interface_unregister_file_receiving(transfer_id, status, size=0, error_message=None):
    if proxy():
        return proxy().callRemote(
            'unregister_file_receiving',
            transfer_id,
            status,
            size,
            error_message,
        ).addErrback(proxy_errback)
    lg.warn('transport_proxy is not ready')
    return fail(Exception('transport_proxy is not ready')).addErrback(proxy_errback)


def interface_cancelled_file_sending(host, filename, size=0, description=None, error_message=None):
    if proxy():
        return proxy().callRemote(
            'cancelled_file_sending',
            'proxy',
            host,
            filename,
            size,
            description,
            error_message,
        ).addErrback(proxy_errback)
    lg.warn('transport_proxy is not ready')
    return fail(Exception('transport_proxy is not ready')).addErrback(proxy_errback)



"""

cmd for the hn_reg_round command.


Usage:

python3 cmd_hn_reg_round.py round

"""

import connections
import json
import socks
import sys

__version__ = '0.0.1'

ORIGIN_OF_TIME = 1534716000  # Real Origin: August 20
POS_SLOT_TIME_MIN = 3  # Real world setting?
POS_SLOT_TIME_SEC = POS_SLOT_TIME_MIN*60
MAX_ROUND_SLOTS = 19  # Real world. 19+1 = 20 , 3x20 = 60 (round time)
END_ROUND_SLOTS = 1
# Round time in seconds
ROUND_TIME_SEC = POS_SLOT_TIME_SEC*(MAX_ROUND_SLOTS + END_ROUND_SLOTS)


def hn_reg_round(round: int, pow_height: int = 0):
    s = socks.socksocket()
    s.settimeout(10)
    s.connect(('127.0.0.1', 5658))
    timestamp = ORIGIN_OF_TIME + round*ROUND_TIME_SEC
    # Last param is ip, to get feed of a specific ip, False for all.
    connections.send(s, 'HN_reg_round {} {} {} False'.format(round, timestamp, pow_height))
    res = connections.receive(s)
    return res


if __name__ == '__main__':
    _, round_string = sys.argv
    res_as_dict = hn_reg_round(int(round_string))
    print('Answer (<=0 means fail):')
    print(json.dumps(res_as_dict, indent=2))



#!/usr/bin/python
# my_keys.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (my_keys.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
"""
.. module:: my_keys.

"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 12

#------------------------------------------------------------------------------

import os
import sys
import gc
import base64

#------------------------------------------------------------------------------

if __name__ == '__main__':
    import os.path as _p
    sys.path.insert(0, _p.abspath(_p.join(_p.dirname(_p.abspath(sys.argv[0])), '..')))

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.system import bpio

from bitdust.lib import misc
from bitdust.lib import strng
from bitdust.lib import jsn
from bitdust.lib import utime

from bitdust.system import local_fs

from bitdust.main import settings
from bitdust.main import events
from bitdust.main import listeners

from bitdust.crypt import key
from bitdust.crypt import rsa_key
from bitdust.crypt import hashes

from bitdust.userid import global_id
from bitdust.userid import id_url
from bitdust.userid import my_id

#------------------------------------------------------------------------------

_KnownKeys = {}
_LatestLocalKeyID = -1
_LocalKeysRegistry = {}
_LocalKeysIndex = {}

#------------------------------------------------------------------------------


def init():
    if _Debug:
        lg.out(_DebugLevel, 'my_keys.init')
    scan_local_keys()


def shutdown():
    global _LatestLocalKeyID
    if _Debug:
        lg.out(_DebugLevel, 'my_keys.shutdown')
    known_keys().clear()
    local_keys().clear()
    local_keys_index().clear()
    _LatestLocalKeyID = 0


#------------------------------------------------------------------------------


def key_obj(key_id=None):
    """
    Alias.
    """
    if not key_id:
        return known_keys()
    if key_id not in known_keys():
        new_key_id = latest_key_id(key_id)
        if new_key_id == key_id:
            raise Exception('key %r is not registered' % key_id)
        if new_key_id not in known_keys():
            raise Exception('key %r is not registered' % new_key_id)
        rename_key(key_id, new_key_id)
        key_id = new_key_id
    if known_keys()[key_id] is None:
        if not load_key(key_id):
            raise Exception('key load failed: %s' % key_id)
    return known_keys()[key_id]


def known_keys():
    """
    Returns dictionary with all registered keys by global identifiers.
    Item value can be None which means the key needs to be loaded first from local file.
    """
    global _KnownKeys
    return _KnownKeys


def local_keys():
    """
    Stores local identifiers of the registered keys.
    """
    global _LocalKeysRegistry
    return _LocalKeysRegistry


def local_keys_index():
    """
    Keeps an index of public key part and local key identifier.
    """
    global _LocalKeysIndex
    return _LocalKeysIndex


#------------------------------------------------------------------------------


def is_key_registered(key_id, include_master=True):
    """
    Returns True if this key is known.
    """
    if include_master:
        if key_id == 'master':
            return True
        if key_id == my_id.getGlobalID():
            return True
        if key_id == my_id.getGlobalID(key_alias='master'):
            return True
    if key_id in known_keys():
        return True
    new_key_id = latest_key_id(key_id)
    if new_key_id in known_keys():
        rename_key(key_id, new_key_id)
        return True
    check_rename_my_keys(prefix=new_key_id.split('@')[0])
    return key_id in known_keys()


def is_key_private(key_id, include_master=True):
    if not is_key_registered(key_id):
        return False
    if include_master:
        if key_id == 'master':
            return True
        if key_id == my_id.getGlobalID():
            return True
        if key_id == my_id.getGlobalID(key_alias='master'):
            return True
    return not key_obj(key_id).isPublic()


#------------------------------------------------------------------------------


def make_key_id(alias, creator_idurl=None, creator_glob_id=None):
    """
    Every key has a creator, and we include his IDURL in the final key_id string.
    Here is a global unique address to a remote copy of `cat.png` file:

        group_abc$alice@first-machine.com:animals/cat.png#F20160313043757PM

    key_id here is:

        group_abc$alice@first-machine.com

    key alias is `group_abc` and creator IDURL is:

        http://first-machine.com/alice.xml

    By knowing full key_id we can find and connect to the correct node(s)
    who is supporting that resource.
    """
    if not alias:
        alias = 'master'
    if creator_glob_id is not None:
        return '{}${}'.format(alias, creator_glob_id)
    if creator_idurl is None:
        creator_idurl = my_id.getIDURL()
    return global_id.MakeGlobalID(
        idurl=creator_idurl,
        key_alias=alias,
    )


def split_key_id(key_id, as_field=True):
    """
    Return "alias" and "creator" IDURL of that key as a tuple object.
    For example from input string:

        "secret_key_xyz$bob@remote-server.net"

    output will be like that:

        "secret_key_xyz", "http://remote-server.net/bob.xml"
    """
    parts = global_id.ParseGlobalID(key_id)
    if not parts['key_alias'] or not parts['idurl']:
        return None, None
    if as_field:
        return parts['key_alias'], id_url.field(parts['idurl'])
    return parts['key_alias'], strng.to_bin(parts['idurl'])


def is_valid_key_id(global_key_id):
    parts = global_id.ParseGlobalID(global_key_id)
    if not parts['key_alias']:
        lg.warn('no key_alias found in the input')
        return False
    if not parts['idurl']:
        lg.warn('no idurl found in the input')
        return False
    if not misc.ValidKeyAlias(parts['key_alias']):
        lg.warn('invalid key alias in the input')
        return False
    return True


def latest_key_id(key_id):
    """
    Create IDURL object from input key_id and return new key_id (with same key_alias) from that IDURL object.
    This way you can be sure that given key_id is pointing to the correct owner IDURL.
    """
    if not key_id:
        return key_id
    key_id = strng.to_text(key_id)
    if key_id == 'master':
        return my_id.getGlobalID(key_alias='master')
    glob_id = global_id.ParseGlobalID(key_id, as_field=True)
    if not glob_id['idurl']:
        lg.err('invalid key_id: %r' % key_id)
        return key_id
    return global_id.MakeGlobalID(
        idurl=glob_id['idurl'].to_bin(),
        key_alias=glob_id['key_alias'],
    )


def get_creator_idurl(key_id, as_field=True):
    """
    Returns creator IDURL from the key_id.
    """
    _, _, creator_glob_id = key_id.partition('$')
    return global_id.glob2idurl(creator_glob_id, as_field=as_field)


#------------------------------------------------------------------------------


def scan_local_keys(keys_folder=None):
    global _LatestLocalKeyID
    if not keys_folder:
        keys_folder = settings.KeyStoreDir()
    if _Debug:
        lg.out(_DebugLevel, 'my_keys.scan_local_keys will read files from %r' % keys_folder)
    latest_local_key_id_filepath = os.path.join(keys_folder, 'latest_local_key_id')
    latest_local_key_id_src = local_fs.ReadTextFile(latest_local_key_id_filepath)
    if latest_local_key_id_src:
        _LatestLocalKeyID = int(latest_local_key_id_src)
    else:
        _LatestLocalKeyID = 0
    known_keys().clear()
    local_keys().clear()
    local_keys_index().clear()
    count = 0
    unregistered_keys = []
    for key_filename in os.listdir(keys_folder):
        if key_filename == 'latest_local_key_id':
            continue
        key_id = key_filename.replace('.private', '').replace('.public', '')
        if not is_valid_key_id(key_id):
            lg.warn('key_id is not valid: %r' % key_id)
            continue
        key_dict = read_key_file(key_id, keys_folder=keys_folder)
        local_key_id = key_dict.get('local_key_id')
        if local_key_id is None:
            key_dict['key_id'] = key_id
            unregistered_keys.append(key_dict)
            continue
        if _LatestLocalKeyID < local_key_id:
            _LatestLocalKeyID = local_key_id
        local_keys()[local_key_id] = key_id
        known_keys()[key_id] = None
        count += 1
    registered_count = 0
    for key_dict in unregistered_keys:
        key_id = key_dict['key_id']
        if not load_key(key_id, keys_folder=keys_folder):
            continue
        _LatestLocalKeyID += 1
        new_local_key_id = _LatestLocalKeyID
        lg.warn('about to register key %r with local_key_id=%r' % (key_id, new_local_key_id))
        known_keys()[key_id].local_key_id = new_local_key_id
        save_key(key_id, keys_folder=keys_folder)
        registered_count += 1
    unregistered_keys = []
    save_latest_local_key_id(keys_folder=keys_folder)
    if _Debug:
        lg.out(_DebugLevel, '    %d keys found and %d registered' % (count, registered_count))


def read_key_file(key_id, keys_folder=None):
    if not keys_folder:
        keys_folder = settings.KeyStoreDir()
    key_filepath = os.path.join(keys_folder, '%s.private' % key_id)
    is_private = True
    if not os.path.exists(key_filepath):
        key_filepath = os.path.join(keys_folder, '%s.public' % key_id)
        is_private = False
    key_raw = local_fs.ReadTextFile(key_filepath)
    if not key_raw:
        lg.warn('failed reading key from %r' % key_filepath)
        return None
    key_raw_strip = key_raw.strip()
    try:
        if key_raw_strip.startswith('{') and key_raw_strip.endswith('}'):
            key_dict = jsn.loads_text(key_raw_strip)
        else:
            key_dict = {
                'label': key_id,
                'is_private': is_private,
                'body': key_raw_strip,
                'local_key_id': None,
                'need_to_convert': True,
                'active': True,
            }
    except:
        lg.exc()
        return None
    return key_dict


def load_key(key_id, keys_folder=None):
    global _LatestLocalKeyID
    if not is_valid_key_id(key_id):
        lg.warn('key is not valid: %r' % key_id)
        return False
    key_dict = read_key_file(key_id, keys_folder=keys_folder)
    try:
        key_object = rsa_key.RSAKey()
        key_object.fromDict(key_dict)
    except:
        lg.exc()
        return False
    if not key_object.isPublic():
        if not validate_key(key_object):
            lg.warn('validation failed for: %r' % key_id)
            return False
    known_keys()[key_id] = key_object
    if key_dict.get('need_to_convert'):
        save_key(key_id, keys_folder=keys_folder)
        lg.info('key %r format converted to JSON' % key_id)
    else:
        if key_object.local_key_id is not None:
            if _LatestLocalKeyID < key_object.local_key_id:
                _LatestLocalKeyID = key_object.local_key_id
                save_latest_local_key_id(keys_folder=keys_folder)
            local_keys()[key_object.local_key_id] = key_id
            local_keys_index()[key_object.toPublicString()] = key_object.local_key_id
            if _Debug:
                lg.out(_DebugLevel, 'my_keys.load_key %r  label=%r  active=%r  is_private=%r  local_key_id=%r  from %s' % (key_id, key_object.label, key_object.active, not key_object.isPublic(), key_object.local_key_id, keys_folder))
        else:
            lg.warn('for key %r local_key_id was not set' % key_id)
    events.send('key-loaded', data=dict(
        key_id=key_id,
        label=key_object.label,
        key_size=key_object.size(),
    ))
    snapshot = make_key_info(
        key_object=key_object,
        key_id=key_id,
        event='key-loaded',
        include_private=False,
        include_local_id=True,
        include_signature=True,
        include_label=True,
        include_state=True,
    )
    listeners.push_snapshot('key', snap_id=key_id, data=snapshot)
    return True


def save_key(key_id, keys_folder=None):
    key_object = known_keys()[key_id]
    if key_object is None:
        lg.warn('can not save key %s because it is not loaded yet' % key_id)
        return False
    if key_object.local_key_id is None:
        raise Exception('local key id was not set for %r' % key_id)
    if not keys_folder:
        keys_folder = settings.KeyStoreDir()
    if key_object.isPublic():
        key_filepath = os.path.join(keys_folder, key_id + '.public')
        key_dict = key_object.toDict(include_private=False)
        key_string = jsn.dumps(key_dict, indent=1, separators=(',', ':'))
    else:
        key_filepath = os.path.join(keys_folder, key_id + '.private')
        key_dict = key_object.toDict(include_private=True)
        key_string = jsn.dumps(key_dict, indent=1, separators=(',', ':'))
    if not bpio.WriteTextFile(key_filepath, key_string):
        lg.warn('failed saving key %r to %r' % (key_id, key_filepath))
        return False
    local_keys()[key_object.local_key_id] = key_id
    local_keys_index()[key_object.toPublicString()] = key_object.local_key_id
    if _Debug:
        lg.out(_DebugLevel, 'my_keys.save_key stored key %r with local_key_id=%r in %r' % (key_id, key_object.local_key_id, key_filepath))
    return True


def save_keys_local(keys_folder=None):
    if not keys_folder:
        keys_folder = settings.KeyStoreDir()
    if _Debug:
        lg.out(_DebugLevel, 'my_keys.save_keys_local will store all known keys in %s' % keys_folder)
    count = 0
    for key_id in known_keys().keys():
        if save_key(key_id, keys_folder=keys_folder):
            count += 1
    if _Debug:
        lg.out(_DebugLevel, '    %d keys saved' % count)
    return count


def save_latest_local_key_id(keys_folder=None):
    global _LatestLocalKeyID
    if not keys_folder:
        keys_folder = settings.KeyStoreDir()
    latest_local_key_id_filepath = os.path.join(keys_folder, 'latest_local_key_id')
    if not bpio.WriteTextFile(latest_local_key_id_filepath, '{}'.format(_LatestLocalKeyID)):
        lg.warn('failed saving latest_local_key_id to %r' % latest_local_key_id_filepath)
        return False
    return True


#------------------------------------------------------------------------------


def generate_key(key_id, label='', active=True, key_size=4096, keys_folder=None):
    global _LatestLocalKeyID
    key_id = latest_key_id(key_id)
    if is_key_registered(key_id):
        lg.warn('key %r already registered' % key_id)
        return None
    if not label:
        label = 'key%s' % utime.make_timestamp()
    if _Debug:
        lg.out(_DebugLevel, 'my_keys.generate_key %r of %d bits, label=%r' % (key_id, key_size, label))
    _LatestLocalKeyID += 1
    save_latest_local_key_id(keys_folder=keys_folder)
    key_object = rsa_key.RSAKey()
    key_object.generate(key_size)
    key_object.label = label
    key_object.active = active
    key_object.local_key_id = _LatestLocalKeyID
    known_keys()[key_id] = key_object
    if _Debug:
        lg.out(_DebugLevel, '    key %r generated' % key_id)
    if not keys_folder:
        keys_folder = settings.KeyStoreDir()
    save_key(key_id, keys_folder=keys_folder)
    events.send('key-generated', data=dict(
        key_id=key_id,
        label=label,
        key_size=key_size,
    ))
    snapshot = make_key_info(
        key_object=key_object,
        key_id=key_id,
        event='key-generated',
        include_private=False,
        include_local_id=True,
        include_signature=True,
        include_label=True,
        include_state=True,
    )
    listeners.push_snapshot('key', snap_id=key_id, data=snapshot)
    return key_object


def register_key(key_id, key_object_or_string, label='', active=True, keys_folder=None):
    global _LatestLocalKeyID
    key_id = latest_key_id(key_id)
    if is_key_registered(key_id):
        lg.warn('key %s already registered' % key_id)
        return None
    if not keys_folder:
        keys_folder = settings.KeyStoreDir()
    if not label:
        label = 'key%s' % utime.make_timestamp()
    if strng.is_string(key_object_or_string):
        key_object_or_string = strng.to_bin(key_object_or_string)
        if _Debug:
            lg.out(_DebugLevel, 'my_keys.register_key %r from %d bytes openssh_input_string' % (key_id, len(key_object_or_string)))
        key_object = unserialize_key_to_object(key_object_or_string)
        if not key_object:
            lg.warn('invalid openssh string, unserialize_key_to_object() failed')
            return None
    else:
        if _Debug:
            lg.out(_DebugLevel, 'my_keys.register_key %r from object' % key_id)
        key_object = key_object_or_string
        label = key_object.label or label
    known_local_key_id = local_keys_index().get(key_object.toPublicString())
    if known_local_key_id is not None:
        known_key_id = local_keys().get(known_local_key_id)
        if known_key_id is not None:
            known_key_id = latest_key_id(known_key_id)
            if known_key_id != key_id:
                raise Exception('must not register same key with local_key_id=%r twice with different key_id: %r ~ %r' % (
                    known_local_key_id,
                    known_key_id,
                    key_id,
                ))
    new_local_key_id = known_local_key_id
    if new_local_key_id is None:
        _LatestLocalKeyID += 1
        save_latest_local_key_id(keys_folder=keys_folder)
        new_local_key_id = _LatestLocalKeyID
    key_object.local_key_id = new_local_key_id
    key_object.label = label
    key_object.active = active
    known_keys()[key_id] = key_object
    if _Debug:
        lg.out(_DebugLevel, '    key %r registered' % key_id)
    save_key(key_id, keys_folder=keys_folder)
    events.send('key-registered', data=dict(
        key_id=key_id,
        local_key_id=new_local_key_id,
        label=label,
        key_size=key_object.size(),
    ))
    snapshot = make_key_info(
        key_object=key_object,
        key_id=key_id,
        event='key-registered',
        include_private=False,
        include_local_id=True,
        include_signature=True,
        include_label=True,
        include_state=True,
    )
    listeners.push_snapshot('key', snap_id=key_id, data=snapshot)
    return key_object


def erase_key(key_id, keys_folder=None):
    key_id = latest_key_id(key_id)
    if not is_key_registered(key_id):
        lg.warn('key %s is not registered' % key_id)
        return False
    if not keys_folder:
        keys_folder = settings.KeyStoreDir()
    if key_obj(key_id).isPublic():
        key_filepath = os.path.join(keys_folder, key_id + '.public')
        is_private = False
    else:
        key_filepath = os.path.join(keys_folder, key_id + '.private')
        is_private = True
    try:
        os.remove(key_filepath)
    except:
        lg.exc()
        return False
    k_obj = known_keys().pop(key_id)
    erased_local_key_id = k_obj.local_key_id
    local_keys().pop(k_obj.local_key_id, None)
    local_keys_index().pop(k_obj.toPublicString(), None)
    gc.collect()
    if _Debug:
        lg.out(_DebugLevel, '    key %s removed, file %s deleted' % (key_id, key_filepath))
    events.send('key-erased', data=dict(key_id=key_id, local_key_id=erased_local_key_id, is_private=is_private))
    snapshot = make_key_info(
        key_object=None,
        key_id=key_id,
        local_id=erased_local_key_id,
        event='key-erased',
        include_private=False,
        include_local_id=True,
        include_signature=True,
        include_label=True,
        include_state=True,
    )
    listeners.push_snapshot('key', snap_id=key_id, deleted=True, data=snapshot)
    return True


def validate_key(key_object):
    sample_data = strng.to_bin(base64.b64encode(os.urandom(256)))
    sample_hash_base = hashes.sha1(sample_data, hexdigest=True)
    sample_signature = key_object.sign(sample_hash_base)
    is_valid = key_object.verify(sample_signature, sample_hash_base)
    if not is_valid:
        if _Debug:
            lg.err('validate_key FAILED')
            lg.out(_DebugLevel, 'public=%r' % key_object.toPublicString())
            lg.out(_DebugLevel, 'signature=%r' % sample_signature)
            lg.out(_DebugLevel, 'hash_base=%r' % sample_hash_base)
            lg.out(_DebugLevel, 'data=%r' % sample_data)
    return is_valid


def rename_key(current_key_id, new_key_id, keys_folder=None):
    if not keys_folder:
        keys_folder = settings.KeyStoreDir()
    if current_key_id not in known_keys():
        lg.warn('key %s is not found' % current_key_id)
        return False
    if key_obj(current_key_id).isPublic():
        current_key_filepath = os.path.join(keys_folder, current_key_id + '.public')
        new_key_filepath = os.path.join(keys_folder, new_key_id + '.public')
        is_private = False
    else:
        current_key_filepath = os.path.join(keys_folder, current_key_id + '.private')
        new_key_filepath = os.path.join(keys_folder, new_key_id + '.private')
        is_private = True
    is_signed = key_obj(current_key_id).isSigned()
    try:
        os.rename(current_key_filepath, new_key_filepath)
    except:
        lg.exc()
        return False
    key_object = known_keys().pop(current_key_id)
    known_keys()[new_key_id] = key_object
    local_keys()[key_object.local_key_id] = new_key_id
    if is_signed:
        sign_key(
            key_id=new_key_id,
            keys_folder=keys_folder,
            ignore_shared_keys=True,
            save=False,
        )
    save_key(new_key_id, keys_folder=keys_folder)
    gc.collect()
    if _Debug:
        lg.out(_DebugLevel, 'my_keys.rename_key   key %s renamed to %s' % (current_key_id, new_key_id))
    events.send('key-renamed', data=dict(old_key_id=current_key_id, new_key_id=new_key_id, is_private=is_private))
    return True


def sign_key(key_id, keys_folder=None, ignore_shared_keys=False, save=True):
    key_id = latest_key_id(strng.to_text(key_id))
    if not is_key_registered(key_id):
        lg.warn('key %s is not found' % key_id)
        return False
    if not keys_folder:
        keys_folder = settings.KeyStoreDir()
    key_object = known_keys()[key_id]
    if key_object.signed:
        if key_object.signed[1] != key.MyPublicKey():
            if ignore_shared_keys:
                if _Debug:
                    lg.dbg(_DebugLevel, 'skip generating signature for shared key: %r' % key_id)
                return True
            raise Exception('must not generate and overwrite existing signature for shared key: %r' % key_id)
    signed_key_info = make_key_info(
        key_object=key_object,
        key_id=key_id,
        include_private=not key_object.isPublic(),
        generate_signature=True,
    )
    key_object.signed = (
        signed_key_info['signature'],
        signed_key_info['signature_pubkey'],
    )
    known_keys()[key_id] = key_object
    if save:
        save_key(key_id, keys_folder=keys_folder)
    events.send('key-signed', data=dict(
        key_id=key_id,
        label=key_object.label,
        key_size=key_object.size(),
    ))
    snapshot = make_key_info(
        key_object=key_object,
        key_id=key_id,
        event='key-signed',
        include_private=False,
        include_local_id=True,
        include_signature=True,
        include_label=True,
        include_state=True,
    )
    listeners.push_snapshot('key', snap_id=key_id, data=snapshot)
    return key_object


#------------------------------------------------------------------------------


def sign(key_id, inp):
    """
    Sign some ``inp`` string with given key.
    This will call PyCrypto method ``Crypto.PublicKey.RSA.sign``.
    Returns byte string.
    """
    key_id = latest_key_id(key_id)
    if not is_key_registered(key_id):
        raise Exception('key %s is not registered' % key_id)
    if known_keys()[key_id] is None:
        if not load_key(key_id):
            raise Exception('key load failed: %s' % key_id)
    key_object = known_keys()[key_id]
    result = key_object.sign(inp)
    return result


def verify(key_id, hashcode, signature):
    """
    Verify signature, this calls function ``Crypto.PublicKey.RSA.verify`` to verify.

    :param key_id: private key id to be used
    :param hashcode: input data to verify, we use method ``Hash`` to prepare that.
    :param signature: string with signature to verify.

    Return True if signature is correct, otherwise False.
    """
    key_id = latest_key_id(key_id)
    if not is_key_registered(key_id):
        raise Exception('key %s is not registered' % key_id)
    if known_keys()[key_id] is None:
        if not load_key(key_id):
            raise Exception('key load failed: %s' % key_id)
    key_object = known_keys()[key_id]
    result = key_object.verify(signature, hashcode)
    return result


#------------------------------------------------------------------------------


def encrypt(key_id, inp):
    """
    Encrypt ``inp`` string using given private key ID.

    :param key_id: private key id to be used
    :param inp: raw binary input string to be encrypted

    Return encrypted string.
    """
    if key_id == 'master':  # master
        if _Debug:
            lg.out(_DebugLevel, 'my_keys.encrypt  payload of %d bytes using my "master" key alias' % len(inp))
        return key.EncryptLocalPublicKey(inp)
    if key_id == my_id.getGlobalID():  # user@host.org
        if _Debug:
            lg.out(_DebugLevel, 'my_keys.encrypt  payload of %d bytes using my "master" key, short format' % len(inp))
        return key.EncryptLocalPublicKey(inp)
    if key_id == my_id.getGlobalID(key_alias='master'):  # master$user@host.org
        if _Debug:
            lg.out(_DebugLevel, 'my_keys.encrypt  payload of %d bytes using my "master" key, full format' % len(inp))
        return key.EncryptLocalPublicKey(inp)
    key_id = latest_key_id(key_id)
    if not is_key_registered(key_id):
        raise Exception('key %s is not registered' % key_id)
    if known_keys()[key_id] is None:
        if not load_key(key_id):
            raise Exception('key load failed: %s' % key_id)
    key_object = known_keys()[key_id]
    if _Debug:
        lg.out(_DebugLevel, 'my_keys.encrypt  payload of %d bytes with key %s' % (len(inp), key_id))
    result = key_object.encrypt(inp)
    return result


def decrypt(key_id, inp):
    """
    Decrypt ``inp`` string with given private key.

    :param key_id: private key id to be used
    :param inp: input string with encrypted data

    Return decrypted string or raise exception.
    """
    if key_id == 'master':  # master
        if _Debug:
            lg.out(_DebugLevel, 'my_keys.decrypt  payload of %d bytes using my "master" key alias' % len(inp))
        return key.DecryptLocalPrivateKey(inp)
    if key_id == my_id.getGlobalID(key_alias='master'):  # master$user@host.org
        if _Debug:
            lg.out(_DebugLevel, 'my_keys.decrypt  payload of %d bytes using my "master" key, full format' % len(inp))
        return key.DecryptLocalPrivateKey(inp)
    if key_id == my_id.getGlobalID():  # user@host.org
        if _Debug:
            lg.out(_DebugLevel, 'my_keys.decrypt  payload of %d bytes using my "master" key, short format' % len(inp))
        return key.DecryptLocalPrivateKey(inp)
    key_id = latest_key_id(key_id)
    if not is_key_registered(key_id):
        raise Exception('key %s is not registered' % key_id)
    if known_keys()[key_id] is None:
        if not load_key(key_id):
            raise Exception('key load failed: %s' % key_id)
    key_object = known_keys()[key_id]
    if _Debug:
        lg.out(_DebugLevel, 'my_keys.decrypt  payload of %d bytes with registered key %s' % (len(inp), key_id))
    result = key_object.decrypt(inp)
    return result


#------------------------------------------------------------------------------


def serialize_key(key_id):
    key_id = latest_key_id(key_id)
    if not is_key_registered(key_id):
        raise Exception('key %s is not registered' % key_id)
    if known_keys()[key_id] is None:
        if not load_key(key_id):
            raise Exception('key load failed: %s' % key_id)
    key_object = known_keys()[key_id]
    return key_object.toPrivateString()


def unserialize_key_to_object(raw_string):
    try:
        key_object = rsa_key.RSAKey()
        key_object.fromString(raw_string)
    except:
        lg.exc()
        return None
    return key_object


#------------------------------------------------------------------------------


def get_public_key_raw(key_id):
    kobj = key_obj(key_id)
    return kobj.toPublicString()


def get_private_key_raw(key_id):
    kobj = key_obj(key_id)
    if kobj.isPublic():
        raise ValueError('not a private key')
    return kobj.toPrivateString()


def get_label(key_id):
    """
    Returns known label for given key.
    """
    key_id = strng.to_text(key_id)
    if not is_key_registered(key_id):
        return ''
    return key_obj(key_id).label


def get_local_key_id(key_id):
    key_id = strng.to_text(key_id)
    if key_id == 'master':
        return 0
    if key_id == my_id.getGlobalID():
        return 0
    if key_id == my_id.getGlobalID(key_alias='master'):
        return 0
    if not is_key_registered(key_id, include_master=False):
        return None
    return key_obj(key_id).local_key_id


def get_local_key(local_key_id):
    if local_key_id == 0:
        return my_id.getGlobalID('master')
    return local_keys().get(local_key_id)


#------------------------------------------------------------------------------


def is_active(key_id):
    """
    Returns True if given key has "active" state. If key is not set to "active" state, certain parts of the software should not use it.
    """
    key_id = strng.to_text(key_id)
    if not is_key_registered(key_id):
        return None
    return key_obj(key_id).active


def set_active(key_id, active=True):
    key_id = strng.to_text(key_id)
    if not is_key_registered(key_id):
        return
    key_obj(key_id).active = active


#------------------------------------------------------------------------------


def make_master_key_info(include_private=False):
    r = {
        'key_id': my_id.getGlobalID(key_alias='master'),
        'alias': 'master',
        'label': my_id.getGlobalID(key_alias='master'),
        'active': True,
        'creator': my_id.getIDURL(),
        'is_public': key.MyPrivateKeyObject().isPublic(),  # 'fingerprint': str(key.MyPrivateKeyObject().fingerprint()),
        # 'type': str(key.MyPrivateKeyObject().type()),
        # 'ssh_type': str(key.MyPrivateKeyObject().sshType()),
        'public': strng.to_text(key.MyPrivateKeyObject().toPublicString()),
        'include_private': include_private,
    }
    r['private'] = None
    if include_private:
        r['private'] = strng.to_text(key.MyPrivateKeyObject().toPrivateString())
    if hasattr(key.MyPrivateKeyObject(), 'size'):
        r['size'] = strng.to_text(key.MyPrivateKeyObject().size())
    else:
        r['size'] = '0'
    return r


def make_key_info(
    key_object, key_id=None, key_alias=None, creator_idurl=None, include_private=False, generate_signature=False, include_signature=False, include_local_id=False, include_label=True, include_state=False, event=None, local_id=None
):
    if key_id:
        key_id = latest_key_id(key_id)
        key_alias, creator_idurl = split_key_id(key_id)
    else:
        key_id = make_key_id(alias=key_alias, creator_idurl=id_url.field(creator_idurl))
    r = {
        'key_id': key_id,
        'alias': key_alias,
        'creator': creator_idurl,
        'public': strng.to_text(key_object.toPublicString()) if key_object else None,
        'private': None,
        'include_private': include_private,
    }
    if event:
        r['event'] = event
    if include_label:
        r['label'] = key_object.label if key_object else ''
    if include_state:
        r['active'] = key_object.active if key_object else True
    if key_object and key_object.isPublic():
        r['is_public'] = True
        if include_private:
            raise Exception('this key contains only public component')
    else:
        r['is_public'] = not include_private
        if include_private:
            r['private'] = strng.to_text(key_object.toPrivateString()) if key_object else None
    if key_object and hasattr(key_object, 'size'):
        r['size'] = strng.to_text(key_object.size())
    else:
        r['size'] = '0'
    if include_local_id:
        if local_id:
            r['local_key_id'] = local_id
        else:
            r['local_key_id'] = getattr(key_object, 'local_key_id', None) if key_object else None
    if key_object and generate_signature:
        r = sign_key_info(r)
    else:
        if include_signature and key_object and key_object.isSigned():
            r['signature'] = key_object.signed[0]
            r['signature_pubkey'] = key_object.signed[1]
    return r


def get_key_info(key_id, include_private=False, include_signature=False, generate_signature=False, include_label=True, include_state=False):
    """
    Returns dictionary with full key info or raise an Exception.
    """
    key_id = strng.to_text(key_id)
    if key_id == 'master' or key_id == my_id.getGlobalID(key_alias='master') or key_id == my_id.getGlobalID():
        return make_master_key_info(include_private=include_private)
    key_id = latest_key_id(key_id)
    key_alias, creator_idurl = split_key_id(key_id)
    if not key_alias or not creator_idurl:
        raise Exception('incorrect key_id format: %s' % key_id)
    if not is_key_registered(key_id):
        key_id = make_key_id(
            alias=key_alias,
            creator_idurl=creator_idurl,
        )
    if not is_key_registered(key_id):
        raise Exception('key %s is not registered' % key_id)
    if known_keys()[key_id] is None:
        if not load_key(key_id):
            raise Exception('key load failed: %s' % key_id)
    key_object = known_keys()[key_id]
    key_info = make_key_info(
        key_object=key_object,
        key_id=key_id,
        include_private=include_private,
        include_signature=include_signature,
        generate_signature=generate_signature,
        include_label=include_label,
        include_state=include_state,
    )
    return key_info


def read_key_info(key_json):
    try:
        key_id = strng.to_text(key_json['key_id'])
        include_private = bool(key_json['include_private'])
        if include_private or key_json.get('private'):
            raw_openssh_string = strng.to_text(key_json['private'])
        else:
            raw_openssh_string = strng.to_text(key_json['public'])
        key_object = unserialize_key_to_object(raw_openssh_string)
        if not key_object:
            raise Exception('unserialize failed')
        key_object.label = strng.to_text(key_json.get('label', ''))
        key_object.active = key_json.get('active', True)
        if 'signature' in key_json and 'signature_pubkey' in key_json:
            key_object.signed = (
                key_json['signature'],
                key_json['signature_pubkey'],
            )
    except:
        lg.exc()
        raise Exception('failed reading key info')
    return latest_key_id(key_id), key_object


#------------------------------------------------------------------------------


def sign_key_info(key_info):
    key_info['signature_pubkey'] = key.MyPublicKey()
    hash_items = []
    for field in [
        'alias',
        'public',
        'signature_pubkey',
    ]:
        hash_items.append(strng.to_text(key_info[field]))
    hash_text = '-'.join(hash_items)
    if _Debug:
        lg.dbg(_DebugLevel, hash_text)
    hash_bin = key.Hash(strng.to_bin(hash_text))
    key_info['signature'] = strng.to_text(key.Sign(hash_bin))
    return key_info


def verify_key_info_signature(key_info):
    if 'signature' not in key_info or 'signature_pubkey' not in key_info:
        lg.warn('signature was not found in the key info')
        return False
    hash_items = []
    for field in [
        'alias',
        'public',
        'signature_pubkey',
    ]:
        hash_items.append(strng.to_text(key_info[field]))
    hash_text = '-'.join(hash_items)
    if _Debug:
        lg.dbg(_DebugLevel, hash_text)
    hash_bin = key.Hash(strng.to_bin(hash_text))
    signature_bin = strng.to_bin(key_info['signature'])
    result = key.VerifySignature(key_info['signature_pubkey'], hash_bin, signature_bin)
    return result


#------------------------------------------------------------------------------


def check_rename_my_keys(prefix=None):
    """
    Make sure all my keys have correct names according to known latest identities I have cached.
    For every key checks corresponding IDURL info and decides to rename it if key owner's identity was rotated.
    """
    keys_to_be_renamed = {}
    for key_id in list(known_keys().keys()):
        if prefix:
            if not key_id.startswith(prefix):
                continue
        key_glob_id = global_id.ParseGlobalID(key_id)
        owner_idurl = key_glob_id['idurl']
        if not owner_idurl.is_latest():
            keys_to_be_renamed[key_id] = global_id.MakeGlobalID(
                idurl=owner_idurl.to_bin(),
                key_alias=key_glob_id['key_alias'],
            )
    if _Debug:
        lg.args(_DebugLevel, keys_to_be_renamed=len(keys_to_be_renamed))
    for current_key_id, new_key_id in keys_to_be_renamed.items():
        rename_key(current_key_id, new_key_id)


#------------------------------------------------------------------------------


def populate_keys():
    for key_id, key_object in known_keys().items():
        listeners.push_snapshot('key', snap_id=key_id, data=make_key_info(
            key_object=key_object,
            key_id=key_id,
            event=None,
            include_private=False,
            include_local_id=True,
            include_signature=True,
            include_label=True,
            include_state=True,
        ))


#------------------------------------------------------------------------------

if __name__ == '__main__':
    bpio.init()
    lg.set_debug_level(18)
    settings.init()
    init()
    import pprint
    pprint.pprint(local_keys_index())
    pprint.pprint(local_keys())
    print(get_key_info('master$recalx@seed.bitdust.io'))
    pprint.pprint(local_keys_index())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2011-2013 Codernity (http://codernity.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import absolute_import
from CodernityDB.hash_index import UniqueHashIndex, HashIndex
from CodernityDB.sharded_index import ShardedIndex
from CodernityDB.index import IndexPreconditionsException

from random import getrandbits
import uuid


class IU_ShardedUniqueHashIndex(ShardedIndex):

    custom_header = """import uuid
from random import getrandbits
from CodernityDB.sharded_index import ShardedIndex
"""

    def __init__(self, db_path, name, *args, **kwargs):
        if kwargs.get('sh_nums', 0) > 255:
            raise IndexPreconditionsException("Too many shards")
        kwargs['ind_class'] = UniqueHashIndex
        super(IU_ShardedUniqueHashIndex, self).__init__(db_path,
                                                        name, *args, **kwargs)
        self.patchers.append(self.wrap_insert_id_index)

    @staticmethod
    def wrap_insert_id_index(db_obj, ind_obj, clean=False):
        def _insert_id_index(_rev, data):
            """
            Performs insert on **id** index.
            """
            _id, value = db_obj.id_ind.make_key_value(data)  # may be improved
            trg_shard = _id[:2]
            storage = db_obj.id_ind.shards_r[trg_shard].storage
            start, size = storage.insert(value)
            db_obj.id_ind.insert(_id, _rev, start, size)
            return _id
        if not clean:
            if hasattr(db_obj, '_insert_id_index_orig'):
                raise IndexPreconditionsException(
                    "Already patched, something went wrong")
            setattr(db_obj, "_insert_id_index_orig", db_obj._insert_id_index)
            setattr(db_obj, "_insert_id_index", _insert_id_index)
        else:
            setattr(db_obj, "_insert_id_index", db_obj._insert_id_index_orig)
            delattr(db_obj, "_insert_id_index_orig")

    def create_key(self):
        h = uuid.UUID(int=getrandbits(128), version=4).hex
        trg = self.last_used + 1
        if trg >= self.sh_nums:
            trg = 0
        self.last_used = trg
        h = '%02x%30s' % (trg, h[2:])
        return h

    def delete(self, key, *args, **kwargs):
        trg_shard = key[:2]
        op = self.shards_r[trg_shard]
        return op.delete(key, *args, **kwargs)

    def update(self, key, *args, **kwargs):
        trg_shard = key[:2]
        self.last_used = int(trg_shard, 16)
        op = self.shards_r[trg_shard]
        return op.update(key, *args, **kwargs)

    def insert(self, key, *args, **kwargs):
        trg_shard = key[:2]  # in most cases it's in create_key BUT not always
        self.last_used = int(key[:2], 16)
        op = self.shards_r[trg_shard]
        return op.insert(key, *args, **kwargs)

    def get(self, key, *args, **kwargs):
        trg_shard = key[:2]
        self.last_used = int(trg_shard, 16)
        op = self.shards_r[trg_shard]
        return op.get(key, *args, **kwargs)


class ShardedUniqueHashIndex(IU_ShardedUniqueHashIndex):

    # allow unique hash to be used directly
    custom_header = 'from CodernityDB.sharded_hash import IU_ShardedUniqueHashIndex'

    pass


class IU_ShardedHashIndex(ShardedIndex):

    custom_header = """from CodernityDB.sharded_index import ShardedIndex"""

    def __init__(self, db_path, name, *args, **kwargs):
        kwargs['ind_class'] = HashIndex
        super(IU_ShardedHashIndex, self).__init__(db_path, name, *
                                                  args, **kwargs)

    def calculate_shard(self, key):
        """
        Must be implemented. It has to return shard to be used by key

        :param key: key
        :returns: target shard
        :rtype: int
        """
        raise NotImplementedError()

    def delete(self, doc_id, key, *args, **kwargs):
        trg_shard = self.calculate_shard(key)
        op = self.shards_r[trg_shard]
        return op.delete(doc_id, key, *args, **kwargs)

    def insert(self, doc_id, key, *args, **kwargs):
        trg_shard = self.calculate_shard(key)
        op = self.shards_r[trg_shard]
        return op.insert(doc_id, key, *args, **kwargs)

    def update(self, doc_id, key, *args, **kwargs):
        trg_shard = self.calculate_shard(key)
        op = self.shards_r[trg_shard]
        return op.insert(doc_id, key, *args, **kwargs)

    def get(self, key, *args, **kwargs):
        trg_shard = self.calculate_shard(key)
        op = self.shards_r[trg_shard]
        return op.get(key, *args, **kwargs)


class ShardedHashIndex(IU_ShardedHashIndex):
    pass



# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (__init__.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
"""
..

module:: chat
"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2011-2013 Codernity (http://codernity.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from threading import RLock

from bitdust_forks.CodernityDB.env import cdb_environment

cdb_environment['mode'] = "threads"
cdb_environment['rlock_obj'] = RLock


from .database_safe_shared import SafeDatabase


class ThreadSafeDatabase(SafeDatabase):

    """
    Thread safe version of CodernityDB that uses several lock objects,
    on different methods / different indexes etc. It's completely different
    implementation of locking than SuperThreadSafe one.
    """
    pass



#!/usr/bin/python
# eccmap.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (eccmap.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: eccmap.

This object holds one error correcting code map.
When creating an object we give a filename to load from or a map name to load from memory.

The spec says for each datablock which parity blocks XOR it.
As we read in each datablock we XOR all the parity blocks for it.
Probably best to read in all the datablocks at once, and do all
the parities one time.  If we did one datablock after another
the parity blocks would need to be active all the time.

If we XOR integers, we have a byte order issue probably, though
maybe not since they all stay the same order, whatever that is.

Some machines are 64-bit. Would be fun to make use of that if we have it.

    >>> 2^1  # ^ is XOR operator
    3

Parity packets have to be a little longer so they can hold
parities on signatures of datapackets.
So we can reconstruct those signatures.

Numbering is from 0 to 63 for range of 64.
"""

from __future__ import absolute_import
from __future__ import print_function
from six.moves import range
from io import open

import os
import re
import traceback

#------------------------------------------------------------------------------

_Debug = False

#------------------------------------------------------------------------------

__suppliers2eccmap = {
    2: 'ecc/2x2',
    4: 'ecc/4x4',
    7: 'ecc/7x7',
    13: 'ecc/13x13',
    18: 'ecc/18x18',
    26: 'ecc/26x26',
    64: 'ecc/64x64',
}

__suppliers_numbers = sorted(__suppliers2eccmap.keys())

__eccmap2suppliers = {
    'ecc/2x2': 2,
    'ecc/4x4': 4,
    'ecc/7x7': 7,
    'ecc/13x13': 13,
    'ecc/18x18': 18,
    'ecc/26x26': 26,
    'ecc/64x64': 64,
}

__eccmap_names = list(__eccmap2suppliers.keys())
__eccmap_names.sort()

__eccmaps = {
    'ecc/2x2': [[1], [0]],
    'ecc/4x4': [[1, 2, 3], [0, 2], [0, 3], [0, 1]],
    'ecc/7x7': [[3, 4, 6], [0, 4, 5], [1, 5, 6], [0, 2, 6], [0, 1, 3], [1, 2, 4], [2, 3, 5]],
    'ecc/13x13': [[1, 4, 8, 12], [5, 8, 9, 11], [3, 7, 10, 11], [0, 4, 6, 9], [2, 3, 6, 12], [0, 1, 6, 10], [1, 3, 7, 9], [2, 5, 8, 12], [2, 4, 7, 11], [0, 1, 3, 5, 12], [6, 7, 8], [2, 5, 9, 10], [0, 4, 10, 11]],
    'ecc/18x18': [
        [5, 7, 11, 16, 17],
        [2, 9, 11, 13, 17],
        [5, 8, 9, 13, 15],
        [0, 1, 4, 6, 10],
        [2, 3, 12, 13, 14],
        [6, 8, 13, 17],
        [2, 5, 10, 12],
        [3, 10, 11, 14],
        [0, 1, 3, 4, 5, 6, 7, 9, 10, 11, 13, 14, 15, 16, 17],
        [0, 1, 12, 14],
        [5, 6, 8, 14, 16],
        [0, 4, 7, 9],
        [2, 4, 7, 8],
        [3, 4, 6, 11, 15],
        [0, 10, 15, 16],
        [1, 2, 17],
        [3, 8, 12, 15],
        [1, 7, 9, 12, 16],
    ],
    'ecc/26x26': [
        [1, 8, 11, 16, 19, 21],
        [3, 6, 8, 17, 23],
        [6, 7, 11, 17, 21, 25],
        [0, 10, 13, 14, 21],
        [5, 9, 10, 18, 22],
        [12, 13, 17, 20, 21, 22],
        [1, 2, 9, 13],
        [2, 3, 5, 9, 20, 22],
        [0, 6, 9, 12, 15, 25],
        [2, 7, 14, 15, 16, 24],
        [2, 5, 6, 11, 15, 16, 18, 19, 23],
        [2, 10, 12, 13, 14, 20, 23],
        [0, 3, 4, 11, 19],
        [0, 1, 4, 18, 19, 20, 23, 25],
        [1, 5, 7, 11, 20, 21, 25],
        [1, 4, 16, 17, 18],
        [2, 4, 11, 22, 24],
        [5, 12, 13, 14, 16, 24],
        [3, 7, 10, 20, 22, 24, 25],
        [0, 8, 10, 12, 17],
        [0, 8, 9, 17, 19, 22, 25],
        [4, 5, 15, 16, 22],
        [6, 8, 12, 14, 15, 18, 23],
        [1, 3, 7, 13, 19, 24],
        [0, 3, 4, 7, 14, 15, 21, 23],
        [6, 8, 9, 10, 18, 24],
    ],
    'ecc/64x64': [
        [5, 17, 18, 31, 39, 47, 55, 58],
        [0, 3, 4, 25, 27, 32, 34, 48, 53, 56, 63],
        [10, 11, 17, 18, 25, 32, 36, 40, 45, 51],
        [1, 21, 23, 27, 30, 35, 43, 47, 62],
        [2, 19, 20, 21, 28, 29, 37, 38, 40, 55, 56, 62],
        [15, 17, 19, 20, 31, 45, 46, 54, 57, 63],
        [19, 20, 30, 36, 46, 47, 52, 62],
        [2, 5, 16, 18, 19, 37, 48, 55],
        [1, 2, 7, 12, 13, 20, 26, 28, 48, 55],
        [0, 1, 15, 21, 24, 33, 36, 41, 56, 62],
        [19, 20, 28, 30, 43, 45, 52, 57, 59],
        [2, 6, 12, 20, 34, 58, 61, 63],
        [5, 6, 13, 15, 25, 34, 36, 40, 42, 43, 50, 51, 55, 61, 62],
        [21, 22, 23, 34, 39, 41, 43, 45, 49, 52, 53, 58],
        [0, 12, 17, 19, 28, 57, 58, 63],
        [8, 18, 25, 29, 34, 49, 52, 53, 56, 62],
        [3, 6, 19, 23, 35, 39, 40, 43, 49, 54, 57],
        [2, 3, 8, 9, 30, 31, 47, 54, 58, 62],
        [0, 8, 14, 24, 28, 33, 36, 47, 52, 58],
        [8, 10, 13, 22, 25, 27, 32, 35, 40, 51, 56],
        [2, 14, 16, 17, 26, 27, 29, 31, 43, 46, 54, 56],
        [22, 25, 37, 41, 45, 52, 61],
        [5, 9, 13, 32, 46, 50, 54, 62],
        [0, 4, 5, 10, 15, 16, 26, 36, 37, 48, 50],
        [13, 14, 20, 21, 40, 42, 55, 60],
        [1, 2, 13, 15, 16, 19, 26, 30, 37, 42, 48, 50, 59],
        [4, 10, 11, 18, 28, 30, 44, 45, 46, 60, 63],
        [2, 6, 16, 22, 24, 38, 41, 53, 59],
        [6, 15, 21, 23, 26, 29, 32, 34, 35, 36, 38, 43, 51, 54, 60],
        [13, 24, 32, 33, 34, 41, 46, 52, 58, 61],
        [1, 10, 23, 24, 27, 29, 40, 41, 61],
        [4, 5, 6, 10, 14, 42, 44, 48, 51, 53, 61],
        [0, 5, 7, 15, 49, 50],
        [8, 29, 35, 36, 43, 47, 51, 60, 62],
        [7, 12, 15, 21, 22, 27, 31, 33, 57, 60],
        [5, 16, 18, 24, 26, 33, 38, 44, 46, 53, 56, 57, 61],
        [1, 3, 4, 9, 24, 27, 31, 39, 50, 51, 54, 58],
        [12, 18, 22, 23, 27, 35, 36, 44, 60, 63],
        [0, 12, 17, 20, 32, 35, 37, 50, 53, 59],
        [8, 11, 14, 16, 22, 24, 35, 36, 41, 42, 44, 46, 57],
        [14, 23, 30, 33, 34, 38, 42, 44, 46, 48, 54],
        [9, 14, 27, 31, 33, 35, 49, 51, 52, 54],
        [3, 8, 11, 12, 14, 30, 32, 34, 48, 56, 62],
        [7, 9, 29, 44, 46, 58],
        [6, 18, 21, 26, 28, 39, 40, 45, 47, 55, 58, 63],
        [4, 17, 21, 26, 30, 34, 54, 61],
        [0, 5, 6, 10, 23, 29, 39, 55, 60],
        [7, 9, 10, 11, 12, 18, 25, 26, 29, 37, 38, 39, 42, 45, 49],
        [6, 7, 17, 27, 33, 56, 59, 60],
        [1, 3, 9, 14, 20, 28, 42, 47, 57, 63],
        [11, 17, 23, 25, 39, 41, 45, 53, 56, 57, 60, 61, 63],
        [4, 8, 12, 16, 19, 28, 31, 32, 47],
        [2, 4, 22, 23, 26, 39, 41, 42, 51, 59],
        [0, 3, 9, 13, 25, 40, 43],
        [0, 9, 10, 16, 22, 47, 53, 55],
        [1, 3, 4, 7, 13, 20, 21, 25, 49, 50],
        [6, 12, 15, 16, 17, 29, 33, 38, 48, 50, 55, 57, 59],
        [1, 15, 24, 28, 37, 40, 42, 52],
        [1, 4, 7, 13, 14, 30, 38, 59],
        [11, 31, 33, 37, 44, 49, 51, 52],
        [8, 11, 24, 31, 32, 35, 50, 53, 59, 63],
        [3, 8, 11, 18, 22, 38, 44, 49],
        [7, 9, 10, 19, 37, 41, 44, 45, 49, 60, 61],
        [2, 3, 5, 7, 11, 38, 39, 43, 48, 59],
    ],
}

__correctable_errors = {
    64: 10,
    26: 6,
    18: 5,
    13: 4,
    7: 3,
    4: 2,
    2: 1,
}

__fire_hire_errors = {
    64: 5,
    26: 3,
    18: 2,
    13: 2,
    7: 2,
    4: 1,
    2: 1,
}

#------------------------------------------------------------------------------

CurrentMap = None

#------------------------------------------------------------------------------


def init():
    """
    Firstly need to call that method, it will set the current map - always keep only one map in memory.
    This module can be imported inside a thread - another map will be created in memory.
    Should have no conflicts with main ecc map. Once the thread is closed - the new map will be destroyed.
    """
    global CurrentMap
    if CurrentMap is not None:
        del CurrentMap
        CurrentMap = None
    CurrentMap = eccmap(CurrentName())


def shutdown():
    """
    This clear the current map from memory.
    """
    global CurrentMap
    del CurrentMap
    CurrentMap = None


#------------------------------------------------------------------------------


def DefaultName():
    """
    This is a wrapper for ``settings.DefaultEccMapName``.
    """
    from bitdust.main import settings
    return settings.DefaultEccMapName()


def CurrentName():
    """
    Should return a ecc map name from current suppliers number - taken from user settings.
    """
    from bitdust.main import settings
    return GetEccMapName(settings.getSuppliersNumberDesired())


def Current():
    """
    Return current map stored in memory, if not yet created - call ``init`` method above.
    """
    global CurrentMap
    if CurrentMap is None:
        init()
    return CurrentMap


def Update():
    """
    Regenerate current map stored in memory from scratch - based on getSuppliersNumberDesired()
    """
    global CurrentMap
    if CurrentMap is not None:
        del CurrentMap
        CurrentMap = None
    CurrentMap = eccmap(CurrentName())


def SuppliersNumbers():
    """
    Return a list of valid suppliers numbers.

    This is: [2, 4, 7, 13, 18, 26, 64]
    """
    global __suppliers_numbers
    return __suppliers_numbers


def EccMapNames():
    """
    Return a list of valid ecc names.
    """
    global __eccmap_names
    return __eccmap_names


def GetEccMapName(suppliers_number):
    """
    Return a ecc map name for given suppliers number or ``DefaultName()``.
    """
    global __suppliers2eccmap
    return __suppliers2eccmap[suppliers_number]


def GetPossibleSuppliersCount():
    global __suppliers2eccmap
    return __suppliers2eccmap.keys()


def GetEccMapSuppliersNumber(eccmapname):
    """
    Reverse method, return a suppliers number for that map.
    """
    global __eccmap2suppliers
    return int(__eccmap2suppliers[eccmapname])


def GetEccMapData(name):
    """
    This return a matrix of that ecc map.

    You can see this is in the top of the file.
    """
    global __eccmaps
    return __eccmaps.get(name, None)


def GetCorrectableErrors(suppliers_number):
    """
    For every map we have different amount of "fixable" errors.
    """
    global __correctable_errors
    return __correctable_errors[suppliers_number]


def GetFireHireErrors(suppliers_number):
    """
    For every map we have different critical amount of dead suppliers.
    """
    global __fire_hire_errors
    return __fire_hire_errors[suppliers_number]


def GetCorrectablePercent(suppliers_number):
    global __correctable_errors
    return 100.0*(float(__correctable_errors[suppliers_number])/float(suppliers_number))


#------------------------------------------------------------------------------


def ReadTextFile(filename):
    """
    Read text file and return its content.

    Also replace line endings: \r\n with \n - convert to Linux file format.
    """
    if not os.path.isfile(filename):
        return ''
    if not os.access(filename, os.R_OK):
        return ''
    try:
        fil = open(filename, 'r')
        data = fil.read()
        fil.close()
        # Windows/Linux trouble with text files
        return data.replace('\r\n', '\n')
    except:
        # lg.exc()
        traceback.print_exc()
    return ''


#------------------------------------------------------------------------------


class eccmap:
    """
    A class to do many operations with ecc map.
    """
    def __init__(self, filename='', suppliers_number=None):
        if filename != '':
            self.name = filename  # sometimes we will just want the name
            self.suppliers_number = GetEccMapSuppliersNumber(self.name)
        else:
            self.suppliers_number = suppliers_number
            self.name = GetEccMapName(self.suppliers_number)
        self.CorrectableErrors = self.CalcCorrectableErrors(self.name)
        self.ParityToData = []  # index with Parity number and output is Data numbers for that parity
        self.DataToParity = []  # index with Data number and output is Parities that use that Data number
        self.datasegments = 0  #
        self.paritysegments = 0  #
        self.type = 0  # 0 is data+parity on same nodes, 1 is different
        self.from_memory(self.name)
        self.convert()

    def __repr__(self):
        return '%s' % self.name

    def NumSuppliers(self):
        """
        Return a number of suppliers for the map.
        """
        return self.suppliers_number

    def DataNeeded(self):
        """
        How many good segments ensures we can fix all.
        """
        return self.datasegments - self.CorrectableErrors

    def CalcCorrectableErrors(self, filename):
        """
        We can fix at least this many errors (probably more for big nums).

        All our codes can handle at least one error.
        """
        basename = os.path.basename(filename)
        CE = 1
        if basename == '64x64':
            CE = 10
        if basename == '26x26':
            CE = 6
        if basename == '18x18':
            CE = 5
        if basename == '13x13':
            CE = 4
        if basename == '7x7':
            CE = 3
        if basename == '4x4':
            CE = 2
        if basename == '2x2':
            CE = 1
        return CE

    def from_memory(self, name):
        """
        Read the constants from memory and take needed matrix on hands.
        """
        # lg.out(6, "eccmap.from_memory with " + name)
        maxdata = 0
        maxparity = 0
        data = GetEccMapData(name)
        self.ParityToData = []
        for PS in data:
            oneset = []
            for datanum in PS:
                try:
                    oneset.append(datanum)
                    if datanum > maxdata:
                        maxdata = datanum
                except (TypeError, ValueError):
                    # lg.out(1, 'eccmap.from_memory ERROR')
                    # lg.exc()
                    traceback.print_exc()
            if oneset:
                self.ParityToData.append(oneset)
                maxparity += 1
        self.paritysegments = maxparity
        self.datasegments = maxdata + 1
        # we only do this type at the moment
        self.type = 0
        # lg.out(6, "   %s with parity=%s data=%s " % (name, self.paritysegments, self.datasegments))

    def loadfromfile(self, fname):
        """
        This is old method, I decide to move all constants into the Python
        code.
        """
        # lg.out(10, "eccmap.loadfromfile with " + fname)
        if os.path.exists(fname):
            filename = fname
        else:
            filename = os.path.join('..', fname)
        maxdata = 0
        maxparity = 0
        s = ReadTextFile(filename)
        for PS in re.split('\]', s):
            good = PS.lstrip('\[\,')
            oneset = []
            for datastring in good.split(','):
                try:
                    datanum = int(datastring)
                    oneset.append(datanum)
                    if datanum > maxdata:
                        maxdata = datanum
                except (TypeError, ValueError):
                    # lg.exc()
                    traceback.print_exc()
            if oneset:
                self.ParityToData.append(oneset)
                maxparity += 1
        self.paritysegments = maxparity
        self.datasegments = maxdata + 1
        self.type = 0  # we only do this type at the moment
        # lg.out(10, "eccmap.loadfromfile  %s  with parity=%s  data=%s " % (filename, self.paritysegments, self.datasegments))

    def convert(self):
        """
        This creates a backward matrix and remember it.
        """
        for datanum in range(self.datasegments):
            self.DataToParity.append([])
        paritynum = 0
        for parity in self.ParityToData:
            for datanum in parity:
                self.DataToParity[datanum].append(paritynum)
            paritynum += 1

    def nodes(self):
        """
        Number of nodes/hosts/sites used for this map.
        """
        if self.type == 0:  # 0 is data+parity on same nodes, 1 is different
            return self.datasegments
        else:
            return self.datasegments + self.paritysegments

    def check(self):
        """
        This is just to test I suppose.
        """
        for emap in self.DataToParity:
            for parity in emap:
                if parity > self.paritysegments:
                    return False
        for emap in self.ParityToData:
            for data in emap:
                if data > self.datasegments:
                    return False
        return True

    def FixableNode(self, NodeMap):
        """
        Case where each node has one parity and one data so same missing.
        """
        DataSegs = [0]*self.datasegments
        ParitySegs = [0]*self.datasegments
        if len(NodeMap) != self.datasegments:
            raise Exception('eccmap.FixableNode not given NodeMap of correct length')
        if self.datasegments != self.paritysegments:
            raise Exception('eccmap.FixableNode only usable if same number of data and parity ')
        for i in range(0, self.datasegments):
            DataSegs[i] = NodeMap[i]
            ParitySegs[i] = NodeMap[i]
        return self.Fixable(DataSegs, ParitySegs)

    def Fixable(self, data_segs, parity_segs):
        """
        Check is reconstruction is possible.

        Lists are 1 for Data and Parity, lists are [0,1,1,1,0...] 0 is
        don't have 1 is have.
        """
        DataSegs = list(data_segs)
        ParitySegs = list(parity_segs)
        stillMissing = 0
        for i in range(self.datasegments):
            if DataSegs[i] != 1:
                stillMissing += 1

        MakingProgress = 1
        # now we test to see if this is fixable
        while MakingProgress == 1 and stillMissing > 0:
            MakingProgress = 0
            for paritynum in range(self.paritysegments):
                if ParitySegs[paritynum] == 1:  # foreach parity
                    Parity = self.ParityToData[paritynum]  # If parity is not missing (so we can use it)
                    missing = 0  # We will see how many of the datas are missing
                    for DataNum in Parity:  # look at all datas that went into this parity
                        if DataSegs[DataNum] == 0:  # if this data is missing
                            missing += 1  # increase count of those missing in this parity
                            lastMissing = DataNum  # keep track of the last missing in case only one is missing
                    if missing == 1:  # if missing exactly 1 of datas in parity
                        MakingProgress = 1  # then we are making progress
                        DataSegs[lastMissing] = 1  # as we could fix lastMissing with current Parity
                        stillMissing -= 1  # so one less stillMissing

        AllFixed = stillMissing == 0  # If nothing else missing we are good
        return AllFixed

    def CanMakeProgress(self, DataSegs, ParitySegs):
        """
        Another method to check if we can do some data reconstruction.

        Lists are 1 for Data and Parity, lists are [0,1,1,1,0...] 0 is
        don't have 1 is have.
        """
        for paritynum in range(self.paritysegments):  # foreach parity
            if ParitySegs[paritynum] == 1:  # If parity is not missing (so we can use it)
                Parity = self.ParityToData[paritynum]
                missing = 0  # We will see how many of the datas are missing
                for DataNum in Parity:  # look at all datas that went into this parity
                    try:
                        if DataSegs[DataNum] != 1:  # if this data is missing
                            missing += 1  # increase count of those missing in this parity
                    except:
                        # lg.exc()
                        traceback.print_exc()
                        return False
                        #     keep track of the last missing in case only one is missing
                if missing == 1:  # if missing exactly 1 of datas in parity, we can fix the data, have work to do
                    return True
            else:  # we're missing this parity, can we rebuild it
                Parity = self.ParityToData[paritynum]
                missing = 0  # We will see how many of the datas are missing
                for DataNum in Parity:  # look at all datas that went into this parity
                    try:
                        if DataSegs[DataNum] != 1:  # if this data is missing
                            missing += 1  # increase count of those missing in this parity
                    except:
                        # lg.exc()
                        traceback.print_exc()
                        return False
                        #     keep track of the last missing in case only one is missing
                if missing == 0:  # if missing none of the data for this parity, we have work to do
                    return True
        return False

    def GetDataFixPath(self, DataSegs, ParitySegs, DataSegNum):
        """
        Given a missing segment number (DataSegNum) and a list of available
        Data Segments and Parity Segments.

        Identify which Parity to use to rebuild the missing Data
        Segment, return the parity segment number and the map of data
        segments in that parity.
        """
        #out(14, 'eccmap.GetDataFixPath %s %s %s' % (str(DataSegNum), str(DataSegs), str(ParitySegs)))
        bestParityNum = -1
        bestParityMap = []
        if DataSegs[DataSegNum] == 1:
            # we have the data segment nothing to fix, unclear why this was called, don't expect this to happen
            #out(12, "eccmap.GetDataFixPath we already have the data segment requested to fix?")
            return bestParityNum, bestParityMap
        for paritynum in range(0, self.paritysegments):
            # If parity is not missing (so we can use it) and contains the missing DataSegNum
            if ParitySegs[paritynum] and (DataSegNum in self.ParityToData[paritynum]):
                Parity = self.ParityToData[paritynum]
                missing = 0  # We will see how many of the datas are missing
                for DataNum in Parity:  # look at all datas that went into this parity
                    if DataSegs[DataNum] == 0:  # if this data is missing
                        missing += 1  # increase count of those missing in this parity
                if missing == 1:  # if missing exactly 1 of datas in parity
                    if (len(bestParityMap) == 0) or (len(Parity) < len(bestParityMap)):
                        bestParityNum = paritynum
                        bestParityMap = Parity
        # if _Debug:
        #     try:
        #         open('/tmp/raid.log', 'a').write(u'GetDataFixPath bestParityNum=%d\n' % bestParityNum)
        #     except:
        #         pass
        return bestParityNum, bestParityMap


#------------------------------------------------------------------------------


def main():
    """
    Tests.
    """
    myecc = eccmap('ecc/4x4')
    print(myecc.ParityToData)
    print(myecc.DataToParity)
    myecc.check()

    print('Do some checks for rebuilding a block')
    dataSegments = [1, 0, 1, 1]
    paritySegments = [1, 1, 1, 0]
    parityNum, parityMap = myecc.GetDataFixPath(dataSegments, paritySegments, 1)
    print(parityNum)  # as of the writing, the best parity was 7
    print(parityMap)
    print(myecc.Fixable(dataSegments, paritySegments))


def main2():
    """
    Another tests.
    """
    myecc = eccmap('ecc/64x64')
    print(myecc.ParityToData)
    print(myecc.DataToParity)
    myecc.check()

    print('Do some checks for rebuilding a block')
    dataSegments = [1]*64
    paritySegments = [1]*64
    dataSegments[2] = 0  # making it so we're missing data segment 2
    parityNum, parityMap = myecc.GetDataFixPath(dataSegments, paritySegments, 2)
    print(parityNum)  # as of the writing, the best parity was 7
    print(parityMap)
    paritySegments[7] = 0  # remove parity 7 and then try it again.
    parityNum, parityMap = myecc.GetDataFixPath(dataSegments, paritySegments, 2)
    print(parityNum)  # should be 11
    print(parityMap)


#------------------------------------------------------------------------------

if __name__ == '__main__':
    main()



#!/usr/bin/python
# service_udp_datagrams.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (service_udp_datagrams.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: service_udp_datagrams
"""

from __future__ import absolute_import
from bitdust.services.local_service import LocalService


def create_service():
    return UDPDatagramsService()


class UDPDatagramsService(LocalService):

    service_name = 'service_udp_datagrams'
    config_path = 'services/udp-datagrams/enabled'
    start_suspended = True

    def dependent_on(self):
        return [
            'service_network',
        ]

    def start(self):
        from bitdust.logs import lg
        from bitdust.lib import udp
        from bitdust.main import settings
        from bitdust.main.config import conf
        udp_port = settings.getUDPPort()
        conf().addConfigNotifier('services/udp-datagrams/udp-port', self._on_udp_port_modified)
        if not udp.proto(udp_port):
            try:
                udp.listen(udp_port)
            except:
                lg.exc()
                return False
        return True

    def stop(self):
        from bitdust.lib import udp
        from bitdust.main import settings
        from bitdust.main.config import conf
        udp_port = settings.getUDPPort()
        if udp.proto(udp_port):
            udp.close(udp_port)
        conf().removeConfigNotifier('services/udp-datagrams/udp-port')
        return True

    def on_suspend(self, *args, **kwargs):
        from bitdust.lib import udp
        from bitdust.main import settings
        udp_port = settings.getUDPPort()
        if udp.proto(udp_port):
            udp.close(udp_port)
        return True

    def on_resume(self, *args, **kwargs):
        from bitdust.logs import lg
        from bitdust.lib import udp
        from bitdust.main import settings
        udp_port = settings.getUDPPort()
        if not udp.proto(udp_port):
            try:
                udp.listen(udp_port)
            except:
                lg.exc()
        return True

    def _on_udp_port_modified(self, path, value, oldvalue, result):
        from bitdust.p2p import network_connector
        network_connector.A('reconnect')
